# -*- coding: utf-8 -*-
"""
Module containing a UTC-based datetime class.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import datetime
import calendar
import math
import operator
import re
import sys
import time
import warnings

import numpy as np
from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning


# based on https://www.myintervals.com/blog/2009/05/20/iso-8601, w/ week 53 fix
_ISO8601_REGEX = re.compile(r"""
    ^
    ([\+-]?\d{4}(?!\d{2}\b))
    ((-?)
     ((0[1-9]|1[0-2])
      (\3([12]\d|0[1-9]|3[01]))?
      |W([0-4]\d|5[0-3])(-?[1-7])?
      |(00[1-9]|0[1-9]\d|[12]\d{2}|3([0-5]\d|6[0-6]))
     )
     ([T\s]
      ((([01]\d|2[0-3])((:?)[0-5]\d)?|24\:?00)([\.,]\d+(?!:))?)?
      (\17[0-5]\d([\.,]\d+)?)?
      ([zZ]|([\+-])([01]\d|2[0-3]):?([0-5]\d)?)?
     )?
    )?
    $
    """, re.VERBOSE)
# Regular expression used in the init function of the UTCDateTime objects which
# is called a lot. Thus pre-compile it.
_YEAR0REGEX = re.compile(r"^(\d{1,3}[-/,])(.*)$")

TIMESTAMP0 = datetime.datetime(1970, 1, 1, 0, 0)

# common attributes
YMDHMS = ('year', 'month', 'day', 'hour', 'minute', 'second')
YJHMS = ('year', 'julday', 'hour', 'minute', 'second')
YMDHMS_FORMAT = "%04d-%02d-%02dT%02d:%02d:%02d"


class UTCDateTime(object):
    """
    A UTC-based datetime object.

    This datetime class is based on the POSIX time, a system for describing
    instants in time, defined as the number of seconds elapsed since midnight
    Coordinated Universal Time (UTC) of Thursday, January 1, 1970. Internally,
    the POSIX time is represented in nanoseconds as an integer, which allows
    higher precision than the default Python :class:`datetime.datetime` class.
    It features the full `ISO8601:2004`_ specification and some additional
    string patterns during object initialization.

    :type args: int, float, str, :class:`datetime.datetime`, optional
    :param args: The creation of a new `UTCDateTime` object depends from the
        given input parameters. All possible options are summarized in the
        `Examples`_ section below.
    :type iso8601: bool or None, optional
    :param iso8601: Enforce/disable `ISO8601:2004`_ mode. Defaults to ``None``
        for auto detection. Works only with a string as first input argument.
    :type strict: bool, optional
    :param strict: If True, Conform to `ISO8601:2004`_ limits on positional
        and keyword arguments. If False, allow hour, minute, second, and
        microsecond values to exceed 23, 59, 59, and 1_000_000 respectively.
    :type precision: int, optional
    :param precision: Sets the precision used by the rich comparison operators.
        Defaults to ``6`` digits after the decimal point. See also `Precision`_
        section below.

    .. versionchanged:: 1.1.0
        UTCDateTime is no longer based on a single floating point value but
        rather an integer representing nanoseconds elapsed since midnight
        Coordinated Universal Time (UTC) of Thursday, January 1, 1970.
        An integer internal representation allows higher precision and more
        predictable behavior than a float representation.

    .. rubric:: Supported Operations

    ``UTCDateTime = UTCDateTime + delta``
        Adds/removes ``delta`` seconds (given as int or float) to/from the
        current ``UTCDateTime`` object and returns a new ``UTCDateTime``
        object.
        See also: :meth:`~obspy.core.utcdatetime.UTCDateTime.__add__`.

    ``delta = UTCDateTime - UTCDateTime``
        Calculates the time difference in seconds between two ``UTCDateTime``
        objects. The time difference is given as float data type and may also
        contain a negative number.
        See also: :meth:`~obspy.core.utcdatetime.UTCDateTime.__sub__`.

    .. rubric:: _`Examples`

    (1) Using a timestamp.

        >>> UTCDateTime(0)
        UTCDateTime(1970, 1, 1, 0, 0)

        >>> UTCDateTime(1240561632)
        UTCDateTime(2009, 4, 24, 8, 27, 12)

        >>> UTCDateTime(1240561632.5)
        UTCDateTime(2009, 4, 24, 8, 27, 12, 500000)

    (2) Using a `ISO8601:2004`_ string. The detection may be enabled/disabled
        using the``iso8601`` parameter, the default is to attempt to
        auto-detect ISO8601 compliant strings.

        * Calendar date representation.

            >>> UTCDateTime("2009-12-31T12:23:34.5")
            UTCDateTime(2009, 12, 31, 12, 23, 34, 500000)

            >>> UTCDateTime("20091231T122334.5")           # compact
            UTCDateTime(2009, 12, 31, 12, 23, 34, 500000)

            >>> UTCDateTime("2009-12-31T12:23:34.5Z")      # w/o time zone
            UTCDateTime(2009, 12, 31, 12, 23, 34, 500000)

            >>> UTCDateTime("2009-12-31T12:23:34+01:15")   # w/ time zone
            UTCDateTime(2009, 12, 31, 11, 8, 34)

        * Ordinal date representation.

            >>> UTCDateTime("2009-365T12:23:34.5")
            UTCDateTime(2009, 12, 31, 12, 23, 34, 500000)

            >>> UTCDateTime("2009365T122334.5")            # compact
            UTCDateTime(2009, 12, 31, 12, 23, 34, 500000)

            >>> UTCDateTime("2009001", iso8601=True)       # enforce ISO8601
            UTCDateTime(2009, 1, 1, 0, 0)

            >>> UTCDateTime("2009360T")                    # compact no time
            UTCDateTime(2009, 12, 26, 0, 0)

        * Week date representation.

            >>> UTCDateTime("2009-W53-7T12:23:34.5")
            UTCDateTime(2010, 1, 3, 12, 23, 34, 500000)

            >>> UTCDateTime("2009W537T122334.5")           # compact
            UTCDateTime(2010, 1, 3, 12, 23, 34, 500000)

            >>> UTCDateTime("2009W011", iso8601=True)      # enforce ISO8601
            UTCDateTime(2008, 12, 29, 0, 0)

        * Specifying time zones.

            >>> UTCDateTime('2019-09-18T+06')  # time zone is UTC+6
            UTCDateTime(2019, 9, 17, 18, 0)

            >>> UTCDateTime('2019-09-18T02-02')  # time zone is UTC-2
            UTCDateTime(2019, 9, 18, 4, 0)

            >>> UTCDateTime('2019-09-18T18:23:10.22-01')  # time zone is UTC-1
            UTCDateTime(2019, 9, 18, 19, 23, 10, 220000)

    (3) Using not ISO8601 compatible strings.

        >>> UTCDateTime("1970-01-01 12:23:34")
        UTCDateTime(1970, 1, 1, 12, 23, 34)

        >>> UTCDateTime("1970,01,01,12:23:34")
        UTCDateTime(1970, 1, 1, 12, 23, 34)

        >>> UTCDateTime("1970,001,12:23:34")
        UTCDateTime(1970, 1, 1, 12, 23, 34)

        >>> UTCDateTime("20090701121212")
        UTCDateTime(2009, 7, 1, 12, 12, 12)

        >>> UTCDateTime("19700101")
        UTCDateTime(1970, 1, 1, 0, 0)

        >>> UTCDateTime("20110818_03:00:00")
        UTCDateTime(2011, 8, 18, 3, 0)

        >>> UTCDateTime("1970/01/17 12:23:34")
        UTCDateTime(1970, 1, 17, 12, 23, 34)

    (4) Using multiple arguments in the following order: `year, month,
        day[, hour[, minute[, second[, microsecond]]]`. The year, month and day
        arguments are required.

        >>> UTCDateTime(1970, 1, 1)
        UTCDateTime(1970, 1, 1, 0, 0)

        >>> UTCDateTime(1970, 1, 1, 12, 23, 34, 123456)
        UTCDateTime(1970, 1, 1, 12, 23, 34, 123456)

    (5) Using the following keyword arguments: `year, month, day, julday, hour,
        minute, second, microsecond`. Either the combination of year, month and
        day, or year and Julian day are required. This is the only input mode
        that supports using hour, minute, or second values above the natural
        limits of 24, 60, 60, respectively.

        >>> UTCDateTime(year=1970, month=1, day=1, minute=15, microsecond=20)
        UTCDateTime(1970, 1, 1, 0, 15, 0, 20)

        >>> UTCDateTime(year=2009, julday=234, hour=14, minute=13)
        UTCDateTime(2009, 8, 22, 14, 13)

    (6) Using a Python :class:`datetime.datetime` object.

        >>> dt = datetime.datetime(2009, 5, 24, 8, 28, 12, 5001)
        >>> UTCDateTime(dt)
        UTCDateTime(2009, 5, 24, 8, 28, 12, 5001)

    (7) Using strict=False the limits of hour, minute, and second become more
        flexible:

        >>> UTCDateTime(year=1970, month=1, day=1, hour=48, strict=False)
        UTCDateTime(1970, 1, 3, 0, 0)

    .. rubric:: _`Precision`

    The :class:`UTCDateTime` class works with a default precision of ``6``
    digits which effects the comparison of date/time values, e.g.:

    >>> dt = UTCDateTime(0)
    >>> dt2 = UTCDateTime(0.00001)
    >>> dt3 = UTCDateTime(0.0000001)
    >>> print(dt.precision)
    6
    >>> dt == dt2  # 5th digit is within current precision
    False
    >>> dt == dt3  # 7th digit will be neglected
    True

    You may change that behavior either by,

    (1) using the ``precision`` keyword during object initialization
        (preferred):

        >>> dt = UTCDateTime(0, precision=4)
        >>> dt2 = UTCDateTime(0.00001, precision=4)
        >>> print(dt.precision)
        4
        >>> dt == dt2
        True

    (2) or by setting the class attribute ``DEFAULT_PRECISION`` to the desired
        precision to affect all new :class:`UTCDateTime` objects
        (not recommended):

        >>> UTCDateTime.DEFAULT_PRECISION = 4
        >>> dt = UTCDateTime(0)
        >>> dt2 = UTCDateTime(0.00001)
        >>> print(dt.precision)
        4
        >>> dt == dt2
        True

        Don't forget to reset ``DEFAULT_PRECISION`` if not needed anymore!

        >>> UTCDateTime.DEFAULT_PRECISION = 6

    .. _ISO8601:2004: https://en.wikipedia.org/wiki/ISO_8601
    """
    DEFAULT_PRECISION = 6
    __slots__ = ('__ns', '__precision', '_initialized', '_has_warned',
                 '__weakref__')

    def __init__(self, *args, ns=None, strict=True, iso8601=None,
                 precision=None, **kwargs):
        """
        Creates a new UTCDateTime object.
        """
        # this is a temporary, it will be removed soon
        self._initialized = False
        self._has_warned = False

        # set default precision
        if precision is None:
            self.__precision = self.DEFAULT_PRECISION
        else:
            self._set_precision(precision)

        # set directly to nanoseconds if given
        if ns is not None:
            self._set_ns(ns)
            return

        # check parameter
        n_args = len(args)
        n_kwargs = len(kwargs)
        if n_args == 0 and n_kwargs == 0:
            # use current date/time if no argument is given
            self._from_timestamp(time.time())
            return
        elif n_args == 1 and n_kwargs == 0:
            value = args[0]
            if isinstance(value, UTCDateTime):
                # ugly workaround to be able to unpickle UTCDateTime objects
                # that were pickled on ObsPy <1.1
                try:
                    self._set_ns(value.__ns)
                except AttributeError:
                    # work around floating point accuracy/rounding issue on
                    # Py3.3, see
                    # https://travis-ci.org/obspy/obspy/jobs/208941376#L751
                    # timestamp is 1251073203.0399999618 so when converting to
                    # integer nanosecond based UTCDateTime this should be
                    # rounded to 1251073203040000 nanoseconds.. but on Py3.3 it
                    # ends up as 1251073203039999, so we manually set
                    # microseconds with correct rounding without artifacts from
                    # floating point precision. see #1664
                    timestamp_seconds = int(value.__dict__['timestamp'])
                    timestamp_microseconds = round(
                        (value.__dict__['timestamp'] % 1.0) * 1e6)
                    dt_ = datetime.datetime.fromtimestamp(
                        timestamp_seconds, tz=datetime.timezone.utc)
                    dt_ = dt_.replace(microsecond=timestamp_microseconds)
                    self._from_datetime(dt_)
                return
            # check types
            # The string instance check is mainly needed to not convert
            # numpy strings as these can be converted to floats on
            # numpy >= 1.14.
            if not isinstance(value, (str, bytes)):
                try:
                    # got a timestamp
                    self._from_timestamp(value.__float__())
                    return
                except Exception:
                    pass
            if isinstance(value, datetime.datetime):
                # got a Python datetime.datetime object
                self._from_datetime(value)
                return
            elif isinstance(value, datetime.date):
                # got a Python datetime.date object
                dt = datetime.datetime(value.year, value.month, value.day)
                self._from_datetime(dt)
                return
            elif isinstance(value, (bytes, str)):
                if not isinstance(value, str):
                    value = value.decode()
                # got a string instance
                value = value.strip()

                # Raising in the case where the leading string is less than 4
                # chars; linked to #2167
                if re.match(_YEAR0REGEX, value):
                    raise ValueError(
                        "'%s' does not start with a 4 digit year" % value)

                # check for ISO8601 date string
                if iso8601 is True or (iso8601 is None and
                                       re.match(_ISO8601_REGEX, value)):
                    try:
                        self._from_iso8601_string(value)
                        return
                    except Exception:
                        # raise here if iso8601 is enforced otherwise fallback
                        # to non iso8601 detection by continuing below
                        if iso8601:
                            raise

                # try to apply some standard patterns
                value = value.replace('T', ' ')
                value = value.replace('_', ' ')
                value = value.replace('-', ' ')
                value = value.replace(':', ' ')
                value = value.replace(',', ' ')
                value = value.replace('/', ' ')
                value = value.replace('Z', ' ')
                value = value.replace('W', ' ')
                # check for ordinal date (julian date)
                parts = value.split(' ')
                # check for patterns
                if len(parts) == 1 and len(value) == 7 and value.isdigit():
                    # looks like an compact ordinal date string
                    pattern = "%Y%j"
                elif len(parts) > 1 and len(parts[1]) == 3 and \
                        parts[1].isdigit():
                    # looks like an ordinal date string
                    value = ''.join(parts)
                    if len(parts) > 2:
                        pattern = "%Y%j%H%M%S"
                    else:
                        pattern = "%Y%j"
                else:
                    # some parts should have 2 digits
                    for i in range(1, min(len(parts), 6)):
                        if len(parts[i]) == 1:
                            parts[i] = '0' + parts[i]
                    value = ''.join(parts)
                    # fill missing elements with zeros
                    value += '0' * (14 - len(value))
                    pattern = "%Y%m%d%H%M%S"
                ms = 0
                if '.' in value:
                    parts = value.split('.')
                    value = parts[0].strip()
                    try:
                        ms = float('.' + parts[1].strip())
                    except Exception:
                        pass
                # all parts should be digits now - here we filter unknown
                # patterns and pass it directly to Python's  datetime.datetime
                if not ''.join(parts).isdigit():
                    dt = datetime.datetime(*args, **kwargs)
                    self._from_datetime(dt)
                    return
                dt = datetime.datetime.strptime(value, pattern)
                dt += datetime.timedelta(seconds=ms)
                self._from_datetime(dt)
                return
        # check for ordinal/julian date kwargs
        if 'julday' in kwargs:
            try:
                int(kwargs['julday'])
            except (ValueError, TypeError):
                msg = "Failed to convert 'julday' to int: {!s}".format(
                    kwargs['julday'])
                raise TypeError(msg)
            if 'year' in kwargs:
                # year given as kwargs
                year = kwargs['year']
            elif n_args == 1:
                # year is first (and only) argument
                year = args[0]
            days_in_year = calendar.isleap(year) and 366 or 365
            if not (1 <= int(kwargs['julday']) <= days_in_year):
                msg = "'julday' out of bounds for year {!s}: {!s}".format(
                    year, kwargs['julday'])
                raise ValueError(msg)
            try:
                temp = "%4d%03d" % (int(year),
                                    int(kwargs['julday']))
                dt = datetime.datetime.strptime(temp, '%Y%j')
            except Exception:
                pass
            else:
                kwargs['month'] = dt.month
                kwargs['day'] = dt.day
                kwargs.pop('julday')

        # check if seconds are given as float value
        if n_args == 6 and isinstance(args[5], float):
            _frac, _sec = math.modf(round(args[5], 6))
            kwargs['second'] = int(_sec)
            kwargs['microsecond'] = int(round(_frac * 1e6))
            args = args[0:5]

        try:  # If a value Error is raised try to allow overflow (see #2222)
            dt = datetime.datetime(*args, **kwargs)
        except ValueError:
            if not strict:
                self._handle_overflow(*args, **kwargs)
            else:
                raise
        else:
            self._from_datetime(dt)

    def _handle_overflow(self, year, month, day, hour=0, minute=0, second=0,
                         microsecond=0):
        """
        Handles setting date if an overflow of usual value limits is detected.
        """
        # Keep track of seconds due to hour, minute, second
        seconds = 0
        seconds += hour * 3600
        seconds += minute * 60
        seconds += second
        # Init UTCDateTime based on year, month, day, add seconds
        utc_base = UTCDateTime(year=year, month=month, day=day)
        # Add seconds and set nanoseconds on self
        self._set_ns((utc_base + seconds + microsecond / 1000000).__ns)

    def _set(self, **kwargs):
        """
        Sets current timestamp using kwargs.
        """
        year = kwargs.get('year', self.year)
        month = kwargs.get('month', self.month)
        day = kwargs.get('day', self.day)
        hour = kwargs.get('hour', self.hour)
        minute = kwargs.get('minute', self.minute)
        second = kwargs.get('second', self.second)
        microsecond = kwargs.get('microsecond', self.microsecond)
        julday = kwargs.get('julday', None)
        if julday:
            self._ns = UTCDateTime(year=year, julday=julday, hour=hour,
                                   minute=minute, second=second,
                                   microsecond=microsecond)._ns
        else:
            self._ns = UTCDateTime(year, month, day, hour, minute,
                                   second, microsecond)._ns

    def _get_ns(self):
        """
        Returns POSIX timestamp as integer nanoseconds.

        This is the internal representation of UTCDateTime objects.

        :rtype: int
        :returns: POSIX timestamp as integer nanoseconds
        """
        return self.__ns

    def _set_ns(self, value):
        """
        Set UTCDateTime object from POSIX timestamp as integer nanoseconds.

        :type value: int
        :param value: POSIX timestamp as integer nanoseconds
        """
        # allow setting numpy integer types..
        if isinstance(value, np.integer):
            value_ = int(value)
            # ..and be paranoid and check that it's still the same value after
            # type casting. Note an assert here since it may not be possible.
            assert value_ == value, (
                "Numpy int "
                f"{value} changed during casting to python int {value_}"
            )
            value = value_
        elif not isinstance(value, int):
            raise TypeError('nanoseconds must be set as int/long type')
        # Store backing value directly to avoid recursive property setter calls
        # when `_ns` is both property name and backing field.
        self.__ns = value
        # flag that this instance has been initialized; any changes will warn
        self._warn_on_change()
        self._initialized = True

    _ns = property(_get_ns, _set_ns)
    ns = property(_get_ns, _set_ns)

    def _from_datetime(self, dt):
        """
        Use Python datetime object to set current time.

        :type dt: :class:`datetime.datetime`
        :param dt: Python datetime object.
        """
        self._set_ns(_datetime_to_ns(dt))

    def _from_timestamp(self, value):
        """
        Use given timestamp to set current time.

        :type value: int, float
        :param value: Timestamp in seconds.
        """
        self._set_ns(int(round(value * 10**9)))

    def _from_iso8601_string(self, value):
        """
        Parses an ISO8601:2004 date time string.
        """
        # remove trailing 'Z'
        value = value.replace('Z', '')
        # split between date and time
        try:
            (date, time) = value.split("T")
        except Exception:
            date = value
            time = ""
        # remove all hyphens in date
        date = date.replace('-', '')
        # remove colons in time
        time = time.replace(':', '')
        # guess date pattern
        length_date = len(date)
        if date.count('W') == 1 and length_date == 8:
            # we got a week date: YYYYWwwD
            # remove week indicator 'W'
            date = date.replace('W', '')
            date_pattern = "%Y%W%w"
            year = int(date[0:4])
            # [Www] is the week number prefixed by the letter 'W', from W01
            # through W53.
            # strpftime %W == Week number of the year (Monday as the first day
            # of the week) as a decimal number [00,53]. All days in a new year
            # preceding the first Monday are considered to be in week 0.
            week = int(date[4:6]) - 1
            # [D] is the weekday number, from 1 through 7, beginning with
            # Monday and ending with Sunday.
            # strpftime %w == Weekday as a decimal number [0(Sunday),6]
            day = int(date[6])
            if day == 7:
                day = 0
            date = "%04d%02d%1d" % (year, week, day)
        elif length_date == 7 and date.isdigit() and value.count('-') != 2:
            # we got a ordinal date: YYYYDDD
            date_pattern = "%Y%j"
        elif length_date == 8 and date.isdigit():
            # we got a calendar date: YYYYMMDD
            date_pattern = "%Y%m%d"
        else:
            raise ValueError("Wrong or incomplete ISO8601:2004 date format")
        # check for time zone information
        # note that the zone designator is the actual offset from UTC and
        # does not include any information on daylight saving time
        if time.count('+') == 1 and '+' in time[-6:]:
            (time, tz) = time.rsplit('+')
            delta = -1
        elif time.count('-') == 1 and '-' in time[-6:]:
            (time, tz) = time.rsplit('-')
            delta = 1
        else:
            delta = 0
        if delta:
            while len(tz) < 3:
                tz += '0'
            delta = delta * (int(tz[0:2]) * 60 * 60 + int(tz[2:]) * 60)
        # split microseconds
        ms = 0
        if '.' in time:
            (time, ms) = time.split(".")
            ms = float('0.' + ms.strip())
        # guess time pattern
        length_time = len(time)
        if length_time == 6 and time.isdigit():
            time_pattern = "%H%M%S"
        elif length_time == 4 and time.isdigit():
            time_pattern = "%H%M"
        elif length_time == 2 and time.isdigit():
            time_pattern = "%H"
        elif length_time == 0:
            time_pattern = ""
        else:
            raise ValueError("Wrong or incomplete ISO8601:2004 time format")
        # parse patterns
        dt = datetime.datetime.strptime(date + 'T' + time,
                                        date_pattern + 'T' + time_pattern)
        # add microseconds and eventually correct time zone
        dt += datetime.timedelta(seconds=float(delta) + ms)
        self._from_datetime(dt)

    def _get_timestamp(self):
        """
        Returns UTC timestamp in seconds.

        :rtype: float
        :return: Timestamp in seconds.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 123456)
        >>> dt.timestamp
        1222864235.123456
        """
        return self.__ns / 1e9

    timestamp = property(_get_timestamp)

    def __float__(self):
        """
        Returns UTC timestamp in seconds.

        :rtype: float
        :return: Timestamp in seconds.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 123456)
        >>> float(dt)
        1222864235.123456
        """
        return self.timestamp

    def _get_datetime(self):
        """
        Returns a Python datetime object.

        :rtype: :class:`datetime.datetime`
        :return: Python datetime object.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.datetime
        datetime.datetime(2008, 10, 1, 12, 30, 35, 45020)
        """
        # datetime.utcfromtimestamp will cut off but not round
        # avoid through adding timedelta - also avoids the year 2038 problem
        rounded_ns = round(self.__ns, self.__precision - 9)
        dt = datetime.timedelta(seconds=rounded_ns // 10**9,
                                microseconds=rounded_ns % 10**9 // 1000)
        try:
            return TIMESTAMP0 + dt
        except OverflowError:
            # for very large future / past dates
            return datetime.datetime.fromtimestamp(
                self.timestamp, tz=datetime.timezone.utc)

    datetime = property(_get_datetime)

    def _get_date(self):
        """
        Returns a Python date object..

        :rtype: :class:`datetime.date`
        :return: Python date object.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.date
        datetime.date(2008, 10, 1)
        """
        return self._get_datetime().date()

    date = property(_get_date)

    def _get_year(self):
        """
        Returns year of the current UTCDateTime object.

        :rtype: int
        :return: Returns year as an integer.

        .. rubric:: Example

        >>> dt = UTCDateTime(2012, 2, 11)
        >>> dt.year
        2012
        """
        return self._get_datetime().year

    def _set_year(self, value):
        """
        Sets year of current UTCDateTime object.

        :param value: Year
        :type value: int
        """
        self._set(year=value)

    year = property(_get_year, _set_year)

    def _get_month(self):
        """
        Returns month as an integer (January is 1, December is 12).

        :rtype: int
        :return: Returns month as an integer, where January is 1 and December
            is 12.

        .. rubric:: Example

        >>> dt = UTCDateTime(2012, 2, 11)
        >>> dt.month
        2
        """
        return self._get_datetime().month

    def _set_month(self, value):
        """
        Sets month of current UTCDateTime object.

        :param value: Month
        :type value: int
        """
        self._set(month=value)

    month = property(_get_month, _set_month)

    def _get_day(self):
        """
        Returns day as an integer.

        :rtype: int
        :return: Returns day as an integer.

        .. rubric:: Example

        >>> dt = UTCDateTime(2012, 2, 11)
        >>> dt.day
        11
        """
        return self._get_datetime().day

    def _set_day(self, value):
        """
        Sets day of current UTCDateTime object.

        :param value: Day
        :type value: int
        """
        self._set(day=value)

    day = property(_get_day, _set_day)

    def _get_weekday(self):
        """
        Return the day of the week as an integer (Monday is 0, Sunday is 6).

        :rtype: int
        :return: Returns day of the week as an integer, where Monday is 0 and
            Sunday is 6.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.weekday
        2
        """
        return self._get_datetime().weekday()

    weekday = property(_get_weekday)

    def _get_time(self):
        """
        Returns a Python time object.

        :rtype: :class:`datetime.time`
        :return: Python time object.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.time
        datetime.time(12, 30, 35, 45020)
        """
        return self._get_datetime().time()

    time = property(_get_time)

    def _get_hour(self):
        """
        Returns hour as an integer.

        :rtype: int
        :return: Returns hour as an integer.

        .. rubric:: Example

        >>> dt = UTCDateTime(2012, 2, 11, 10, 11, 12)
        >>> dt.hour
        10
        """
        return self._get_datetime().hour

    def _set_hour(self, value):
        """
        Sets hours of current UTCDateTime object.

        :param value: Hours
        :type value: int
        """
        self._set(hour=value)

    hour = property(_get_hour, _set_hour)

    def _get_minute(self):
        """
        Returns minute as an integer.

        :rtype: int
        :return: Returns minute as an integer.

        .. rubric:: Example

        >>> dt = UTCDateTime(2012, 2, 11, 10, 11, 12)
        >>> dt.minute
        11
        """
        return self._get_datetime().minute

    def _set_minute(self, value):
        """
        Sets minutes of current UTCDateTime object.

        :param value: Minutes
        :type value: int
        """
        self._set(minute=value)

    minute = property(_get_minute, _set_minute)

    def _get_second(self):
        """
        Returns seconds as an integer.

        :rtype: int
        :return: Returns seconds as an integer.

        .. rubric:: Example

        >>> dt = UTCDateTime(2012, 2, 11, 10, 11, 12)
        >>> dt.second
        12
        """
        return self._get_datetime().second

    def _set_second(self, value):
        """
        Sets seconds of current UTCDateTime object.

        :param value: Seconds
        :type value: int
        """
        self._set(second=value)

    second = property(_get_second, _set_second)

    def _get_microsecond(self):
        """
        Returns microseconds as an integer.

        :rtype: int
        :return: Returns microseconds as an integer.

        .. rubric:: Example

        >>> dt = UTCDateTime(2012, 2, 11, 10, 11, 12, 345234)
        >>> dt.microsecond
        345234
        """
        ms = int(round(self.__ns % 10**9, self.__precision - 9) // 1000)
        return ms % 1000000

    def _set_microsecond(self, value):
        """
        Sets microseconds of current UTCDateTime object.

        :param value: Microseconds
        :type value: int
        """
        self._set(microsecond=value)

    microsecond = property(_get_microsecond, _set_microsecond)

    def _get_julday(self):
        """
        Returns Julian day as an integer.

        :rtype: int
        :return: Julian day as an integer.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.julday
        275
        """
        return self.utctimetuple().tm_yday

    def _set_julday(self, value):
        """
        Sets Julian day of current UTCDateTime object.

        :param value: Julian day
        :type value: int
        """
        self._set(julday=value)

    julday = property(_get_julday, _set_julday)

    def timetuple(self):
        """
        Return a time.struct_time such as returned by time.localtime().

        :rtype: time.struct_time
        """
        return self.datetime.timetuple()

    def utctimetuple(self):
        """
        Return a time.struct_time of current UTCDateTime object.

        :rtype: time.struct_time
        """
        return self.datetime.utctimetuple()

    def __add__(self, value):
        """
        Adds seconds and microseconds to current UTCDateTime object.

        :type value: int, float
        :param value: Seconds to add
        :rtype: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :return: New UTCDateTime object.

        .. rubric:: Example

        >>> dt = UTCDateTime(1970, 1, 1, 0, 0)
        >>> dt + 2
        UTCDateTime(1970, 1, 1, 0, 0, 2)

        >>> UTCDateTime(1970, 1, 1, 0, 0) + 1.123456
        UTCDateTime(1970, 1, 1, 0, 0, 1, 123456)
        """
        if isinstance(value, datetime.timedelta):
            # see datetime.timedelta.total_seconds
            value = (value.microseconds + (value.seconds + value.days *
                     86400) * 10**6) / 1e6
        elif isinstance(value, UTCDateTime):
            msg = ("unsupported operand type(s) for +: 'UTCDateTime' and "
                   "'UTCDateTime'")
            raise TypeError(msg)
        # need to make sure we don't get e.g. np.float32 singl precision input
        # or worse, because then numpy is in charge of the calculations and
        # numpy 2.0 is not automatically upcasting to avoid precision loss
        # which means we can't keep full precision when converting input
        # seconds to nanoseconds
        value = float(value)
        return UTCDateTime(ns=self.__ns + int(round(value * 1e9)))

    def __sub__(self, value):
        """
        Subtracts seconds and microseconds from current UTCDateTime object.

        :type value: int, float or :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param value: Seconds or UTCDateTime object to subtract. Subtracting an
            UTCDateTime objects results into a relative time span in seconds.
        :rtype: :class:`~obspy.core.utcdatetime.UTCDateTime` or float
        :return: New UTCDateTime object or relative time span in seconds.

        .. rubric:: Example

        >>> dt = UTCDateTime(1970, 1, 2, 0, 0)
        >>> dt - 2
        UTCDateTime(1970, 1, 1, 23, 59, 58)

        >>> UTCDateTime(1970, 1, 2, 0, 0) - 1.123456
        UTCDateTime(1970, 1, 1, 23, 59, 58, 876544)

        >>> UTCDateTime(1970, 1, 2, 0, 0) - UTCDateTime(1970, 1, 1, 0, 0)
        86400.0
        """
        if isinstance(value, UTCDateTime):
            return round((self._ns - value._ns) / 1e9, self.__precision)
        elif isinstance(value, datetime.timedelta):
            # see datetime.timedelta.total_seconds
            value = (value.microseconds + (value.seconds + value.days *
                     86400) * 10**6) / 1e6
        return UTCDateTime(ns=self.__ns - int(round((value * 1e9))))

    def __str__(self):
        """
        Returns ISO8601 string representation from current UTCDateTime object.

        :return: string

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> str(dt)
        '2008-10-01T12:30:35.045020Z'
        """
        dt = self.datetime
        time_str = YMDHMS_FORMAT % tuple(getattr(dt, x) for x in YMDHMS)

        if self.precision > 0:
            ns = round(self.ns, self.precision - 9)
            ns_str = ('%09d' % (ns % 10 ** 9))[:self.precision]
            time_str += ('.' + ns_str)
        return time_str + 'Z'

    def _repr_pretty_(self, p, cycle):  # @UnusedVariable
        p.text(str(self))

    def __unicode__(self):
        """
        Returns ISO8601 unicode representation from current UTCDateTime object.

        :return: string

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.__unicode__()
        '2008-10-01T12:30:35.045020Z'
        """
        return str(self.__str__())

    def _operate(self, other, op_func):
        try:
            if self.__precision != other.__precision:
                msg = ('Comparing UTCDateTime objects of different precision'
                       ' is not defined will raise an Exception in a future'
                       ' version of obspy')
                warnings.warn(msg, ObsPyDeprecationWarning)
            ndigits = min(self.__precision, other.__precision) - 9
            lhs = round(self.__ns, ndigits)
            rhs = round(other.__ns, ndigits)
            return op_func(lhs, rhs)
        except AttributeError:
            try:
                return self._operate(UTCDateTime(other), op_func)
            except TypeError:
                return False

    def __eq__(self, other):
        """
        Rich comparison operator '=='.

        .. rubric: Example

        Comparing two UTCDateTime objects will compare the nanoseconds integers
        rounded to a number of significant digits determined by the precision
        attribute.

        >>> t1 = UTCDateTime(123.000000012)
        >>> t2 = UTCDateTime(123.000000099)
        >>> t1 == t2
        True

        Defining a higher precision changes the behavior of the operator

        >>> t1 = UTCDateTime(123.000000012, precision=9)
        >>> t2 = UTCDateTime(123.000000099, precision=9)
        >>> t1 == t2
        False
        """
        return self._operate(other, operator.eq)

    def __ne__(self, other):
        """
        Rich comparison operator '!='.

        .. rubric: Example

        Comparing two UTCDateTime objects will compare the nanoseconds integers
        rounded to a number of significant digits determined by the precision
        attribute.

        >>> t1 = UTCDateTime(123.000000012)
        >>> t2 = UTCDateTime(123.000000099)
        >>> t1 == t2
        True

        Defining a higher precision changes the behavior of the operator

        >>> t1 = UTCDateTime(123.000000012, precision=9)
        >>> t2 = UTCDateTime(123.000000099, precision=9)
        >>> t1 == t2
        False
        """
        return not self.__eq__(other)

    def __lt__(self, other):
        """
        Rich comparison operator '<'.

        .. rubric: Example

        Comparing two UTCDateTime objects will compare the nanoseconds integers
        rounded to a number of significant digits determined by the precision
        attribute.

        >>> t1 = UTCDateTime(123.000000012)
        >>> t2 = UTCDateTime(123.000000099)
        >>> t1 < t2
        False

        Defining a higher precision changes the behavior of the operator

        >>> t1 = UTCDateTime(123.000000012, precision=9)
        >>> t2 = UTCDateTime(123.000000099, precision=9)
        >>> t1 < t2
        True
        """
        return self._operate(other, operator.lt)

    def __le__(self, other):
        """
        Rich comparison operator '<='.

        .. rubric: Example

        Comparing two UTCDateTime objects will compare the nanoseconds integers
        rounded to a number of significant digits determined by the precision
        attribute.

        >>> t1 = UTCDateTime(123.000000099)
        >>> t2 = UTCDateTime(123.000000012)
        >>> t1 <= t2
        True

        Defining a higher precision changes the behavior of the operator

        >>> t1 = UTCDateTime(123.000000099, precision=9)
        >>> t2 = UTCDateTime(123.000000012, precision=9)
        >>> t1 <= t2
        False
        """
        return self._operate(other, operator.le)

    def __gt__(self, other):
        """
        Rich comparison operator '>'.

        .. rubric: Example

        Comparing two UTCDateTime objects will compare the nanoseconds integers
        rounded to a number of significant digits determined by the precision
        attribute.

        >>> t1 = UTCDateTime(123.000000099)
        >>> t2 = UTCDateTime(123.000000012)
        >>> t1 > t2
        False

        Defining a higher precision changes the behavior of the operator

        >>> t1 = UTCDateTime(123.000000099, precision=9)
        >>> t2 = UTCDateTime(123.000000012, precision=9)
        >>> t1 > t2
        True
        """
        return self._operate(other, operator.gt)

    def __ge__(self, other):
        """
        Rich comparison operator '>='.

        .. rubric: Example

        Comparing two UTCDateTime objects will compare the nanoseconds integers
        rounded to a number of significant digits determined by the precision
        attribute.

        >>> t1 = UTCDateTime(123.000000012)
        >>> t2 = UTCDateTime(123.000000099)
        >>> t1 >= t2
        True

        Defining a higher precision changes the behavior of the operator

        >>> t1 = UTCDateTime(123.000000012, precision=9)
        >>> t2 = UTCDateTime(123.000000099, precision=9)
        >>> t1 >= t2
        False
        """
        return self._operate(other, operator.ge)

    def __repr__(self):
        """
        Returns a representation of UTCDatetime object.
        """
        return 'UTCDateTime' + self.datetime.__repr__()[17:]

    def __abs__(self):
        """
        Returns absolute timestamp value of the current UTCDateTime object.
        """
        # needed for unittest.assertAlmostEqual tests on Linux
        return abs(self.timestamp)

    def __hash__(self):
        """
        An object is hashable if it has a hash value which never changes
        during its lifetime. As an UTCDateTime object may change over time,
        it's not hashable. Use the :meth:`~UTCDateTime.datetime()` method to
        generate a :class:`datetime.datetime` object for hashing. But be aware:
        once the UTCDateTime object changes, the hash is not valid anymore.
        """
        # explicitly flag it as unhashable
        return None

    def __getstate__(self):
        """
        Support for pickling UTCDateTime objects.
        """
        return (self.__ns, self.__precision)

    def __setstate__(self, state):
        """
        Support for unpickling UTCDateTime objects.
        """
        self._initialized = False
        self._has_warned = False
        if isinstance(state, tuple):
            self.__ns, self.__precision = state
        elif isinstance(state, dict):
            ns = state.get("_ns", state.get("_UTCDateTime__ns"))
            if ns is None and "timestamp" in state:
                ns = int(round(state["timestamp"] * 10**9))
            if ns is None:
                msg = "Cannot reconstruct UTCDateTime from pickle state"
                raise ValueError(msg)
            precision = state.get(
                "precision",
                state.get("_UTCDateTime__precision", self.DEFAULT_PRECISION))
            self.__ns = int(ns)
            self.__precision = int(precision)
        else:
            msg = f"Unsupported UTCDateTime pickle state type: {type(state)!r}"
            raise TypeError(msg)
        self._initialized = True

    def strftime(self, format):
        """
        Return a string representing the date and time, controlled by an
        explicit format string.

        :type format: str
        :param format: Format string.
        :return: Formatted string representing the date and time.

        Format codes referring to hours, minutes or seconds will see 0 values.
        See methods :meth:`~datetime.datetime.strftime()` and
        :meth:`~datetime.datetime.strptime()` for more information.
        """
        # This is an attempt to get consistent behavior across platforms.
        # See https://bugs.python.org/issue32195
        # and https://bugs.python.org/issue13305
        # This is an issue of glibc implementation differing across platforms,
        # out of control of Python, but we still try to be consistent across
        # all platforms
        if sys.platform.startswith("linux"):
            format = format.replace("%Y", "%04Y")
        return self._get_datetime().strftime(format)

    @staticmethod
    def strptime(date_string, format):
        """
        Return a UTCDateTime corresponding to date_string, parsed according to
        given format.

        :type date_string: str
        :param date_string: Date and time string.
        :type format: str
        :param format: Format string.
        :return: :class:`~obspy.core.utcdatetime.UTCDateTime`

        See methods :meth:`~datetime.datetime.strftime()` and
        :meth:`~datetime.datetime.strptime()` for more information.
        """
        return UTCDateTime(datetime.datetime.strptime(date_string, format))

    def timetz(self):
        """
        Return time object with same hour, minute, second, microsecond, and
        tzinfo attributes. See also method :meth:`datetime.datetime.time()`.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.timetz()
        datetime.time(12, 30, 35, 45020)
        """
        return self._get_datetime().timetz()

    def utcoffset(self):
        """
        Returns None (to stay compatible with :class:`datetime.datetime`)

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.utcoffset()
        """
        return self._get_datetime().utcoffset()

    def dst(self):
        """
        Returns None (to stay compatible with :class:`datetime.datetime`)

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.dst()
        """
        return self._get_datetime().dst()

    def tzname(self):
        """
        Returns None (to stay compatible with :class:`datetime.datetime`)

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.tzname()
        """
        return self._get_datetime().tzname()

    def ctime(self):
        """
        Return a string representing the date and time.

        .. rubric:: Example

        >>> UTCDateTime(2002, 12, 4, 20, 30, 40).ctime()
        'Wed Dec  4 20:30:40 2002'
        """
        return self._get_datetime().ctime()

    def isoweekday(self):
        """
        Return the day of the week as an integer (Monday is 1, Sunday is 7).

        :rtype: int
        :return: Returns day of the week as an integer, where Monday is 1 and
            Sunday is 7.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.isoweekday()
        3
        """
        return self._get_datetime().isoweekday()

    def isocalendar(self):
        """
        Returns a tuple containing (ISO year, ISO week number, ISO weekday).

        :rtype: tuple(int)
        :return: Returns a (named) tuple containing ISO year, ISO week number
            and ISO weekday. Depending on the used Python version it either
            returns a tuple (Py<3.9) or named tuple (Py>=3.9).

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> tuple(dt.isocalendar())
        (2008, 40, 3)
        """
        return self._get_datetime().isocalendar()

    def isoformat(self, sep="T"):
        """
        Return a string representing the date and time in ISO 8601 format.

        :rtype: str
        :return: String representing the date and time in ISO 8601 format like
            YYYY-MM-DDTHH:MM:SS.mmmmmm or, if microsecond is 0,
            YYYY-MM-DDTHH:MM:SS.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> dt.isoformat()
        '2008-10-01T12:30:35.045020'

        >>> dt = UTCDateTime(2008, 10, 1)
        >>> dt.isoformat()
        '2008-10-01T00:00:00'
        """
        return self._get_datetime().isoformat(sep=sep)

    def format_fissures(self):
        """
        Returns string representation for the Earthscope/IRIS Fissures
        protocol.

        :return: string

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> print(dt.format_fissures())
        2008275T123035.0450Z
        """
        return "%04d%03dT%02d%02d%02d.%04dZ" % \
            (self.year, self.julday, self.hour, self.minute, self.second,
             self.microsecond // 100)

    def format_arclink(self):
        """
        Returns string representation for the ArcLink protocol.

        :return: string

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> print(dt.format_arclink())
        2008,10,1,12,30,35,45020
        """
        return "%d,%d,%d,%d,%d,%d,%d" % (self.year, self.month, self.day,
                                         self.hour, self.minute, self.second,
                                         self.microsecond)

    def format_seedlink(self):
        """
        Returns string representation for the SeedLink protocol.

        :return: string

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35.45020)
        >>> print(dt.format_seedlink())
        2008,10,1,12,30,35
        """
        # round seconds down to integer
        seconds = int(float(self.second) + float(self.microsecond) / 1.0e6)
        return "%d,%d,%d,%d,%d,%g" % (self.year, self.month, self.day,
                                      self.hour, self.minute, seconds)

    def format_seed(self, compact=False):
        """
        Returns string representation for a SEED volume.

        :type compact: bool, optional
        :param compact: Delivers a compact SEED date string if enabled. Default
            value is set to False.
        :rtype: str
        :return: Datetime string in the SEED format.

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 10, 1, 12, 30, 35, 45020)
        >>> print(dt.format_seed())
        2008,275,12:30:35.0450

        >>> dt = UTCDateTime(2008, 10, 1, 0, 30, 0, 0)
        >>> print(dt.format_seed(compact=True))
        2008,275,00:30
        """
        if not compact:
            if self.time == datetime.time(0):
                return "%04d,%03d" % (self.year, self.julday)
            return "%04d,%03d,%02d:%02d:%02d.%04d" % (self.year, self.julday,
                                                      self.hour, self.minute,
                                                      self.second,
                                                      self.microsecond // 100)
        temp = "%04d,%03d" % (self.year, self.julday)
        if self.time == datetime.time(0):
            return temp
        temp += ",%02d" % (self.hour)
        if self.microsecond:
            return temp + ":%02d:%02d.%04d" % (self.minute, self.second,
                                               self.microsecond // 100)
        elif self.second:
            return temp + ":%02d:%02d" % (self.minute, self.second)
        elif self.minute:
            return temp + ":%02d" % (self.minute)
        return temp

    def format_iris_web_service(self):
        """
        Returns string representation usable for the EarthScope/IRIS Web
        services.

        :return: string

        .. rubric:: Example

        >>> dt = UTCDateTime(2008, 5, 27, 12, 30, 35, 45020)
        >>> print(dt.format_iris_web_service())
        2008-05-27T12:30:35.045
        """
        return "%04d-%02d-%02dT%02d:%02d:%02d.%03d" % \
            (self.year, self.month, self.day, self.hour, self.minute,
             self.second, self.microsecond // 1000)

    def _get_precision(self):
        """
        Returns precision of current UTCDateTime object.

        :return: int

        .. rubric:: Example

        >>> dt = UTCDateTime()
        >>> dt.precision
        6
        """
        return self.__precision

    def _set_precision(self, value=6):
        """
        Set precision of current UTCDateTime object.

        :type value: int, optional
        :param value: Precision value used by the rich comparison operators.
            Defaults to ``6``.

        .. rubric:: Example

        (1) Default precision

            >>> dt = UTCDateTime()
            >>> dt.precision
            6

        (2) Set precision during initialization of UTCDateTime object.

            >>> dt = UTCDateTime(precision=5)
            >>> dt.precision
            5
        """
        if value > 9:
            msg = 'UTCDateTime precision above 9 is not supported, using 9'
            warnings.warn(msg)
            value = 9
        self.__precision = int(value)
        self._warn_on_change()

    precision = property(_get_precision, _set_precision)

    def _warn_on_change(self):
        if self._initialized and not self._has_warned:
            msg = ('Setting attributes on UTCDateTime instances will raise an'
                   ' Exception in a future version of Obspy.')
            warnings.warn(msg, ObsPyDeprecationWarning)
            # only issue the warning once per object
            self._has_warned = True

    def replace(self, **kwargs):
        """
        Return a new UTCDateTime object with one or more parameters replaced.

        Replace is useful for substituting parameters that depend on other
        parameters (eg hour depends on the current day for meaning). In order
        to replace independent parameters, such as timestamp, ns, or
        precision, simply create a new UTCDateTime instance.

        The following parameters are supported: year, month, day, julday,
        hour, minute second, microsecond. Additionally, the keyword 'strict'
        can be set to False to allow hour, minute, and second to exceed normal
        limits.

        .. rubric:: Example

        (1) Get time of the 15th day of the same month to which a timestamp
            belongs.

            >>> dt = UTCDateTime(999999999)
            >>> dt2 = dt.replace(day=15)
            >>> print(dt2)
            2001-09-15T01:46:39.000000Z


        (2) Determine the day of the week 2 months before Guy Fawkes day.

            >>> dt = UTCDateTime('1605-11-05')
            >>> dt.replace(month=9).weekday
            0
        """
        # check parameters, raise Value error if any are unsupported
        supported_args = set(YMDHMS) | set(YJHMS) | {'microsecond', 'strict'}
        if not set(kwargs).issubset(supported_args):
            unsupported_args = set(kwargs) - supported_args
            msg = ('%s are not supported arguments for replace, supported '
                   'arguments are %s') % (unsupported_args, supported_args)
            raise ValueError(msg)
        # ensure julday is used correctly if used
        if kwargs.get('julday') is not None:
            if 'month' in kwargs or 'day' in kwargs:
                msg = 'If julday is used month and day cannot be used.'
                raise ValueError(msg)
            time_paramters = YJHMS  # use julday

        else:
            time_paramters = YMDHMS  # use month and day
        # get a dict of time parameters to pass to UTCDateTime constructor
        new_dict = {x: getattr(self, x) for x in time_paramters}
        new_dict['microsecond'] = self.microsecond
        new_dict.update(kwargs)
        return UTCDateTime(**new_dict)

    def toordinal(self):
        """
        Return proleptic Gregorian ordinal. January 1 of year 1 is day 1.

        See :meth:`datetime.datetime.toordinal()`.

        :return: int

        .. rubric:: Example

        >>> dt = UTCDateTime(2012, 1, 1)
        >>> dt.toordinal()
        734503
        """
        return self._get_datetime().toordinal()

    @staticmethod
    def now():
        """
        Returns current UTC datetime.
        """
        return UTCDateTime()

    @staticmethod
    def utcnow():
        """
        Returns current UTC datetime.
        """
        return UTCDateTime()

    def _get_hours_after_midnight(self):
        """
        Calculate foating point hours after midnight.

        >>> t = UTCDateTime("2015-09-27T03:16:12.123456Z")
        >>> t._get_hours_after_midnight()
        3.270034293333333
        """
        datetime = self._get_datetime()
        timedelta = (
            datetime -
            datetime.replace(hour=0, minute=0, second=0, microsecond=0))
        return timedelta.total_seconds() / 3600.0

    @property
    def matplotlib_date(self):
        """
        Maplotlib date number representation.

        Useful for plotting on matplotlib time-based axes, like created by e.g.
        :meth:`obspy.core.stream.Stream.plot()`.

        >>> t = UTCDateTime("2009-08-24T00:20:07.700000Z")
        >>> t.matplotlib_date  # doctest: +SKIP
        14480.01397800926

        :rtype: float
        """
        from matplotlib.dates import date2num
        return date2num(self.datetime)


def _datetime_to_ns(dt):
    """
    Use Python datetime object to return equivalent nanoseconds.

    :type dt: :class:`datetime.datetime`
    :param dt: Python datetime object.
    :returns: nanoseconds as an int.
    """
    try:
        td = (dt - TIMESTAMP0)
    except TypeError:
        td = (dt.replace(tzinfo=None) - dt.utcoffset()) - TIMESTAMP0
    # see datetime.timedelta.total_seconds
    return (td.days * 86400 + td.seconds) * 10**9 + td.microseconds * 1000


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import hilbert

from obspy.signal.tf_misfit import plot_tf_gofs


# general constants
tmax = 6.
dt = 0.01
npts = int(tmax / dt + 1)
t = np.linspace(0., tmax, npts)

fmin = .5
fmax = 10
nf = 100

# constants for the signal
A1 = 4.
t1 = 2.
f1 = 2.
phi1 = 0.

# amplitude and phase error
phase_shift = 0.8
amp_fac = 3.

# generate the signal
H1 = (np.sign(t - t1) + 1) / 2
st1 = A1 * (t - t1) * np.exp(-2 * (t - t1))
st1 *= np.cos(2. * np.pi * f1 * (t - t1) + phi1 * np.pi) * H1

# reference signal
st2 = st1.copy()

# generate analytical signal (hilbert transform) and add phase shift
st1p = hilbert(st1)
st1p = np.real(np.abs(st1p) *
               np.exp((np.angle(st1p) + phase_shift * np.pi) * 1j))

# signal with amplitude error
st1a = st1 * amp_fac

plot_tf_gofs(st1a, st2, dt=dt, fmin=fmin, fmax=fmax, show=False)
plot_tf_gofs(st1p, st2, dt=dt, fmin=fmin, fmax=fmax, show=False)

plt.show()



# -*- coding: utf-8 -*-
import numpy as np

from .parser import \
    parse_wtn_frame_to_alsep_words, parse_wth_frame_to_geophone_data


class _WtFrame(object):

    def __init__(self, frame):
        self.data = np.require(frame, dtype=np.int64)
        # Frame header parameters
        self.flag_bit = None
        self.msec_of_year = None
        self.alsep_tracking_station_id = None
        self.alsep_package_id = None
        self.bit_search = None
        self.bit_verify = None
        self.bit_confirm = None
        self.bit_lock = None
        self.bit_input_level = None
        self.original_record_number = None
        self.barker_code = None
        self.barker_code_comp = None
        self.frame_count = None
        self.mode_bit = None

    def parse_header(self):
        # a flag bit indicating usage of computer-generated time code
        self.flag_bit = self.data[0] >> 7

        # time of the year in milliseconds
        self.msec_of_year = np.int64(self.data[0] & 0x7f)
        self.msec_of_year = (self.msec_of_year << 8) + self.data[1]
        self.msec_of_year = (self.msec_of_year << 8) + self.data[2]
        self.msec_of_year = (self.msec_of_year << 8) + self.data[3]
        self.msec_of_year = (self.msec_of_year << 4) + (self.data[4] >> 4)

        # ALSEP tracking station ID
        self.alsep_tracking_station_id = self.data[4] & 0x0f

        # ALSEP package ID
        self.alsep_package_id = self.data[5] >> 5

        # bit synchronizer status
        self.bit_search = self.data[5] & 0x0f
        self.bit_search = self.bit_search >> 4
        self.bit_verify = self.data[5] & 0x08
        self.bit_verify = self.bit_verify >> 3
        self.bit_confirm = self.data[5] & 0x04
        self.bit_confirm = self.bit_confirm >> 2
        self.bit_lock = self.data[5] & 0x02
        self.bit_lock = self.bit_lock >> 1
        self.bit_input_level = self.data[5] & 0x01

        # original 7-track record number (first subframe only)
        self.original_record_number = self.data[6]
        self.original_record_number = \
            (self.original_record_number << 8) + self.data[7]

        # sync pattern - Barker code (11100010010) and its complement
        self.barker_code = self.data[8]
        self.barker_code = (self.barker_code << 3) + (self.data[9] >> 5)
        self.barker_code_comp = self.data[9] & 0x0f
        self.barker_code_comp = \
            (self.barker_code_comp << 7) + (self.data[10] >> 1)

    def is_valid(self):
        if self.alsep_package_id < 1 or self.alsep_package_id > 5:
            return False
        return True

    def dump(self):
        print('--- Frame info ---')
        print('flag_bit: {}'.format(self.flag_bit))
        print('msec of year: {:d}'.format(self.msec_of_year))
        print('ALSEP tracking station ID: {}'.format(
            self.alsep_tracking_station_id))
        print('ALSEP Package ID: {}'.format(self.alsep_package_id))
        print('Bit search: {}'.format(self.bit_search))
        print('Bit verify: {}'.format(self.bit_verify))
        print('Bit confirm: {}'.format(self.bit_confirm))
        print('Bit lock: {}'.format(self.bit_lock))
        print('Bit input level: {}'.format(self.bit_input_level))
        print('Original record number: {}'.format(
            self.original_record_number))
        print('sync pattern barker code: {}'.format(self.barker_code))
        print('sync pattern barker code compliment: {}'.format(
            self.barker_code_comp))
        print('mode bit: {}'.format(self.mode_bit))


class WtnFrame(_WtFrame):
    def __init__(self, frame):
        super(WtnFrame, self).__init__(frame)
        # Extracted ALSEP words
        self.alsep_words = None
        # Parse header and data
        self.parse_header()
        self.parse_data()

    def parse_header(self):
        super(WtnFrame, self).parse_header()

        # frame count
        self.frame_count = self.data[11] >> 1

        # mode bit
        #   - 1 for frame 1 means normal bit rate
        #   - 1 for frame 2 means slow bit rate
        #   - ALSEP ID for frames 3-5 as follows
        #       Station 11 ... 011
        #       Station 12 ... 010
        #       Station 14 ... 110
        #       Station 15 ... 011
        #       Station 16 ... 001
        self.mode_bit = self.data[11] & 0x01

    def parse_data(self):
        self.alsep_words = parse_wtn_frame_to_alsep_words(self.data)

    def dump(self):
        super(WtnFrame, self).dump()
        print('ALSEP words: {}'.format(self.alsep_words))


class WthFrame(_WtFrame):
    def __init__(self, frame):
        super(WthFrame, self).__init__(frame)
        # Extracted Geophone data
        self.geophone = None
        # Parse header and data
        self.parse_header()
        self.parse_data()

    def parse_header(self):
        super(WthFrame, self).parse_header()

    def parse_data(self):
        self.geophone = parse_wth_frame_to_geophone_data(self.data)

    def dump(self):
        super(WthFrame, self).dump()
        print('Geophone: {}'.format(self.geophone))



# -*- coding: utf-8 -*-
import io
import os

from .blockette import Blockette
from ..fields import FixedString, Float, Integer, Loop, VariableString
from ..utils import lookup_code, format_resp


class Blockette041(Blockette):
    """
    Blockette 041: FIR Dictionary Blockette.

    The FIR blockette is used to specify FIR (Finite Impulse Response)
    digital filter coefficients. It is an alternative to blockette [44] when
    specifying FIR filters. The blockette recognizes the various forms of
    filter symmetry and can exploit them to reduce the number of factors
    specified in the blockette. See Response (Coefficients) Blockette [54]
    for more information.
    """

    id = 41
    name = "FIR Dictionary"
    fields = [
        Integer(3, "Response Lookup Key", 4),
        VariableString(4, "Response Name", 1, 25, 'UN_'),
        FixedString(5, "Symmetry Code", 1, 'U'),
        Integer(6, "Signal In Units", 3, xpath=34),
        Integer(7, "Signal Out Units", 3, xpath=34),
        Integer(8, "Number of Factors", 4),
        # REPEAT field 9 for the Number of Factors
        Loop("FIR Coefficient", "Number of Factors", [
            Float(9, "FIR Coefficient", 14, mask='%+1.7e')], flat=True),
    ]

    def parse_seed(self, data, expected_length=0):
        """
        If number of FIR coefficients are larger than maximal blockette size of
        9999 chars a follow up blockette with the same blockette id and
        response lookup key is expected - this is checked here.
        """
        # convert to stream for test issues
        if isinstance(data, bytes):
            expected_length = len(data)
            data = io.BytesIO(data)
        elif isinstance(data, str):
            raise TypeError("Data must be bytes, not string")
        # get current lookup key
        pos = data.tell()
        data.read(7)
        global_lookup_key = int(data.read(4))
        data.seek(pos)
        # read first blockette
        temp = io.BytesIO()
        temp.write(data.read(expected_length))
        # check next blockettes
        while True:
            # save position
            pos = data.tell()
            try:
                blockette_id = int(data.read(3))
            except ValueError:
                break
            if blockette_id != 41:
                # different blockette id -> break
                break
            blockette_length = int(data.read(4))
            lookup_key = int(data.read(4))
            if lookup_key != global_lookup_key:
                # different lookup key -> break
                break
            # ok follow up blockette found - skip some unneeded fields
            self.fields[1].read(data)
            self.fields[2].read(data)
            self.fields[3].read(data)
            self.fields[4].read(data)
            self.fields[5].read(data)
            # remaining length in current blockette
            length = pos - data.tell() + blockette_length
            # read follow up blockette and append it to temporary blockette
            temp.write(data.read(length))
        # reposition file pointer
        data.seek(pos)
        # parse new combined temporary blockette
        temp.seek(0, os.SEEK_END)
        _len = temp.tell()
        temp.seek(0)
        Blockette.parse_seed(self, temp, expected_length=_len)

    def parse_xml(self, xml_doc, *args, **kwargs):
        if self.xseed_version == '1.0':
            xml_doc.find('fir_coefficient').tag = 'FIR_coefficient'
        Blockette.parse_xml(self, xml_doc, *args, **kwargs)

    def get_xml(self, *args, **kwargs):
        xml = Blockette.get_xml(self, *args, **kwargs)
        if self.xseed_version == '1.0':
            xml.find('FIR_coefficient').tag = 'fir_coefficient'
        return xml

    def get_resp(self, station, channel, abbreviations):
        """
        Returns RESP string.
        """
        string = \
            '#\t\t+                     +--------------------------------+' + \
            '                      +\n' + \
            '#\t\t+                     |   FIR response,' + \
            '%6s ch %s   |                      +\n' % (station, channel) + \
            '#\t\t+                     +--------------------------------+' + \
            '                      +\n' + \
            '#\t\t\n' + \
            'B041F05     Symmetry type:                         %s\n' \
            % self.symmetry_code + \
            'B041F06     Response in units lookup:              %s - %s\n' \
            % (lookup_code(abbreviations, 34, 'unit_name',
                           'unit_lookup_code', self.signal_in_units),
               lookup_code(abbreviations, 34, 'unit_description',
                           'unit_lookup_code', self.signal_in_units)) + \
            'B041F07     Response out units lookup:             %s - %s\n' \
            % (lookup_code(abbreviations, 34, 'unit_name', 'unit_lookup_code',
                           self.signal_out_units),
               lookup_code(abbreviations, 34, 'unit_description',
                           'unit_lookup_code', self.signal_out_units)) + \
            'B041F08     Number of numerators:                  %s\n' \
            % self.number_of_factors

        if self.number_of_factors > 1:
            string += '#\t\tNumerator coefficients:\n' + \
                      '#\t\t  i, coefficient\n'
            for _i in range(self.number_of_factors):
                string += 'B041F09    %4s %13s\n' \
                    % (_i, format_resp(self.FIR_coefficient[_i], 6))
        elif self.number_of_factors == 1:
            string += '#\t\tNumerator coefficients:\n' + \
                '#\t\t  i, coefficient\n'
            string += 'B041F09    %4s %13s\n' \
                % (0, format_resp(self.FIR_coefficient, 6))
        string += '#\t\t\n'
        return string.encode()



# -*- coding: utf-8 -*-
"""
Assign ALSEP words to each channel
"""
from .define import FORMAT_ALSEP_PSE_OLD, FORMAT_ALSEP_WTN
from .util import interp


def assign_alsep_words(alsep_word, apollo_station, format_type, frame_count,
                       prev_values):
    data = {}
    # spz (Apollo 11,12,14,15,16) or lsg(Apollo 17)
    if format_type == FORMAT_ALSEP_PSE_OLD and apollo_station != 17:
        spz = [None] * 32
        for i in range(32):
            spz[i] = alsep_word[(i + 1) * 2]
        if prev_values is None:
            spz[0] = spz[1]
        else:
            spz[0] = interp(prev_values[0], prev_values[1], spz[1], spz[2])
        if apollo_station == 15:
            spz[11] = interp(spz[9], spz[10], spz[12], spz[13])
        if apollo_station != 14:
            spz[22] = interp(spz[20], spz[21], spz[23], spz[24])
        spz[27] = interp(spz[25], spz[26], spz[28], spz[29])
        data['spz'] = spz
    elif apollo_station == 17:
        lsg = [None] * 32
        for i in range(32):
            lsg[i] = 511 - alsep_word[(i + 1) * 2]
        if prev_values is None:
            lsg[0] = lsg[1]
        else:
            lsg[0] = interp(prev_values[0], prev_values[1], lsg[1], lsg[2])
        data['lsg'] = lsg
        data['lsg_tide'] = [511 - alsep_word[25]]
        data['lsg_free'] = [511 - alsep_word[27]]
        data['lsg_temp'] = [511 - alsep_word[29]]
    # lpx/lpy/lpz
    lpx = [None] * 4
    lpy = [None] * 4
    lpz = [None] * 4
    for i in range(4):
        lpx[i] = alsep_word[16 * i + 9]
        lpy[i] = alsep_word[16 * i + 11]
        lpz[i] = alsep_word[16 * i + 13]
    data['lpx'] = lpx
    data['lpy'] = lpy
    data['lpz'] = lpz
    # TidalX/TidalY/TidalZ/Inst_temp
    if frame_count % 2 == 0:
        data['tdx'] = [alsep_word[35]]
        data['tdy'] = [alsep_word[37]]
    else:
        data['tdz'] = [alsep_word[35]]
        data['inst_temp'] = [alsep_word[37]]
    # LSM
    if format_type == FORMAT_ALSEP_WTN:
        if apollo_station in [12, 15, 16]:
            lsm = [None] * 6
            lsm[0] = alsep_word[17]
            lsm[1] = alsep_word[19]
            lsm[2] = alsep_word[21]
            lsm[3] = alsep_word[49]
            lsm[4] = alsep_word[51]
            lsm[5] = alsep_word[53]
            data['lsm'] = lsm
            data['lsm_status'] = [alsep_word[5]]
    # House Keeping (HK)
    data['hk'] = [alsep_word[33]]
    # Command Verification (CV)
    if apollo_station == 14:
        data['cv'] = [(alsep_word[5] >> 1)]
    else:
        data['cv'] = [(alsep_word[46] >> 1)]

    return data



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
GCF bindings to python
"""
import ctypes
from io import IOBase

import math
import os
import warnings

import numpy as np

from obspy import UTCDateTime
from obspy.core import Stream, Trace, AttribDict
from obspy.core.util.libnames import _load_cdll

# supported sampling rates outside range 1-250 Hz
# key:value = samplingrate:fractional_time_denominator
_SPS_MAP = {
    0.1: 1,
    0.125: 1,
    0.2: 1,
    0.25: 1,
    0.5: 1,
    400: 8,
    500: 2,
    625: 5,
    800: 16,
    1000: 4,
    1250: 5,
    2000: 8,
    2500: 10,
    4000: 16,
    5000: 20
}
# Unsupported sampling rates in range 1-250 Hz
_SPS_RESERVED = [157, 161, 162, 164, 167, 171, 174, 175, 176, 179, 181, 182,
                 191, 193, 194]
# Valid differential gains
_VALID_GAIN = [-1, 0, 1, 2, 4, 8, 16, 32, 64]
# Min/max data values
_GCF_MIN = -2147483648
_GCF_MAX = 2147483647

# Mapping between error codes set by reader and messages
_ERRORMSG_READ = {
    -1: "data block %d is not a data block",
    1: "stream ID in header of data block %d use extended system ID format",
    2: "stream ID in header of data block %d use double-extended "
       "system ID format",
    3: "unknown compression in data block %d",
    4: "to few/many data samples indicated in block header of data block %d",
    5: "start time of first data sample in data block %d is negative",
    9: "failure to decode header values of data block %d",
    10: "failure to decode data block %d (last data != RIC)",
    11: "1'st first difference != 0 in data block %d",
    21: "failure to decode data block %d (last data != RIC, 1'st first "
        "difference != 0)"
}
_ERROR_WRITE = {
    -2: IOError("failed to write to disc"),
    -1: IOError("failed to open file to write to"),
    1: ValueError("no data or inconsistent headers"),
    2: ValueError("unsupported sampling rate"),
    3: ValueError("bad fractional start time"),
    4: ValueError("unsupported gain"),
    5: ValueError("unsupported instrument type"),
    6: ValueError("too long system_id"),
    }


class _GcfSeg(ctypes.Structure):
    _fields_ = [
      ('streamID', ctypes.c_char * 7),
      ('systemID', ctypes.c_char * 7),
      ('start', ctypes.c_int32),
      ('t_numerator', ctypes.c_int),
      ('t_denominator', ctypes.c_int),
      ('t_leap', ctypes.c_int),
      ('gain', ctypes.c_int),
      ('sysType', ctypes.c_int),
      ('digi', ctypes.c_int),
      ('ttl', ctypes.c_int),
      ('blk', ctypes.c_int),
      ('err', ctypes.c_int),
      ('sps', ctypes.c_int),
      ('sps_denom', ctypes.c_int),
      ('compr', ctypes.c_int),
      ('FIC', ctypes.c_int32),
      ('RIC', ctypes.c_int32),
      ('n_data', ctypes.c_int32),
      ('n_alloc', ctypes.c_int32),
      ('data', ctypes.POINTER(ctypes.c_int32))
    ]


class _GcfFile(ctypes.Structure):
    _fields_ = [
      ('n_blk', ctypes.c_int),
      ('n_seg', ctypes.c_int),
      ('n_alloc', ctypes.c_int),
      ('n_errHead', ctypes.c_int),
      ('n_errData', ctypes.c_int),
      ('seg', ctypes.POINTER(_GcfSeg))
    ]


def compatible_sps(sps):
    """
    Checks if a sampling rate is compatible with the GCF format

    :type sps: float, int
    :param sps: sampling rate to test
    :rtype: bool
    :returns: True if sampling rate is compatible else False
    """
    if not (isinstance(sps, int) or isinstance(sps, float)):
        return False
    if 1 <= sps <= 250:
        if int(sps)-sps != 0 or sps in _SPS_RESERVED:
            return False
    elif sps not in _SPS_MAP:
        return False
    return True


def get_time_denominator(sps):
    """
    Returns the time fractional offset denominator, d, associated with an
    input sampling rate.

    Any data written to a gcf file must have its first sample sampled at a
    fractional time:
        n/d
    where 0 <= n < d

    :type sps: float, int
    :param sps: sampling rate (samples per second)
    :rtype: int
    :returns: fractional offset denominator, if 0 input sampling rate is not
        compatible with implemented GCF format, if 1 first data sample must be
        sampled in integer time
    """
    denom = 0
    if compatible_sps(sps):
        if sps <= 250:
            denom = 1
        else:
            denom = _SPS_MAP[sps]
    return denom


def merge_gcf_stream(st):
    """
    Merges GCF stream (replacing Stream.merge(-1) for headonly=True)

    :type st: :class:`~obspy.core.stream.Stream`
    :param st: GCF Stream object with no data
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream object containing header and data.
    """
    traces = []
    for tr in st:
        delta = tr.stats.delta
        starttime = tr.stats.starttime
        endtime = tr.stats.endtime
        for trace in traces:
            if tr.id == trace.id and delta == trace.stats.delta \
              and not starttime == trace.stats.starttime:
                if 0 < starttime - trace.stats.endtime <= delta:
                    trace.stats.npts += tr.stats.npts
                    break
                elif 0 < trace.stats.starttime - endtime <= delta:
                    trace.stats.starttime = UTCDateTime(starttime)
                    trace.stats.npts += tr.stats.npts
                    break
        else:
            traces.append(tr)
    return Stream(traces=traces)


def _is_gcf(file):
    """
    Checks whether a file is GCF or not.

    :type file: str or file-like object
    :param file: path to GCF file to be checked
    :rtype: bool
    :return: True if a object pointed to by path is a GCF file.
    """
    input_is_gcf = True

    # Load shared library
    gcf_io = _load_cdll("gcf")

    # declare function argument and return types
    gcf_io.free_GcfFile.argtypes = [ctypes.POINTER(_GcfFile)]
    gcf_io.free_GcfFile.restype = None
    # avoid releasing uninitialized memory in free_GcfFile later:
    gcf_io.init_GcfFile.argtypes = [ctypes.POINTER(_GcfFile)]
    gcf_io.init_GcfFile.restype = None

    obj = None
    try:
        # Decode first block:
        obj = _GcfFile()
        # make sure pointers in obj are properly initialized to NULL:
        gcf_io.init_GcfFile(obj)
        ret = _fill_gcf_from_file(file, obj, 3)

        # b_filename = filename.encode('utf-8')
        # ret = gcf_io.read_gcf(fid, obj, 3)
        if ret or (obj.n_errHead and obj.seg[0].err not in (10, 11, 21)) \
                or obj.n_errData:
            input_is_gcf = False
    except Exception:
        input_is_gcf = False
    finally:
        # release allocated memory
        if obj is not None:
            gcf_io.free_GcfFile(obj)

    return input_is_gcf


def _fill_gcf_from_file(file, obj, mode):
    """Read data from `file` into the given `_GcfFile` object `obj`.
    This function is the Python implementation of the C function
    `gcf_io.read_gcf` to allow reading from both file paths
    and file-like objects, and should not be called directly:
    see `_read_gcf` and `_is_gcf` for details.

    :param file: str, Path or file-like object
    :param obj: an empty instance of a _GcfFile` object
    :param mode: the read mode. See `_read_gcf` for details
    """
    # Load shared library
    gcf_io = _load_cdll("gcf")

    # declare function argument and return types
    gcf_io.read_gcf.argtypes = [ctypes.c_char_p,
                                ctypes.POINTER(_GcfFile),
                                ctypes.c_int]
    gcf_io.read_gcf.restype = ctypes.c_int
    gcf_io.init_GcfSeg_for_read.argtypes = [ctypes.POINTER(_GcfSeg),
                                            ctypes.c_int]
    gcf_io.init_GcfSeg_for_read.restype = None
    gcf_io.is_LittleEndian_gcf.argtypes = []
    gcf_io.is_LittleEndian_gcf.restype = ctypes.c_int
    gcf_io.read_gcf_block.argtypes = [ctypes.POINTER(_GcfFile),
                                      ctypes.c_char_p,
                                      ctypes.POINTER(_GcfSeg),
                                      ctypes.c_int,
                                      ctypes.c_int,
                                      ctypes.c_double]
    gcf_io.read_gcf_block.restype = ctypes.c_int
    gcf_io.free_GcfSeg.argtypes = [ctypes.POINTER(_GcfSeg)]
    gcf_io.free_GcfSeg.restype = None
    gcf_io.merge_GcfFile.argtypes = [ctypes.POINTER(_GcfFile),
                                     ctypes.c_int,
                                     ctypes.c_double]
    gcf_io.merge_GcfFile.restype = None

    d = 0  # number of empty blocks
    cur_pos = None
    ret = 0
    fpt = None  # file-like object
    try:
        if isinstance(file, IOBase):
            cur_pos = file.tell()
            fpt = file
        else:
            fpt = open(file, "rb")
    except Exception:
        ret = -1

    # get endianess of current machine
    endian = gcf_io.is_LittleEndian_gcf()

    if not ret:
        b1 = 0
        tol = 1.E-3
        seg = _GcfSeg()
        # initiate and allocate segment
        gcf_io.init_GcfSeg_for_read(seg, mode)

        # adjust mode if nedded
        if mode > 2:
            mode = 2
            b1 = 1

        while True:
            buffer = fpt.read(1024)
            if not buffer:
                break
            elif len(buffer) < 1024:
                ret = -1
                break
            d += gcf_io.read_gcf_block(obj, buffer, seg, mode, endian, tol)
            if b1:
                break

        # free segment
        gcf_io.free_GcfSeg(seg)

        # merge segments if asked for
        if abs(mode) < 2:
            gcf_io.merge_GcfFile(obj, mode, tol)

    if obj.n_blk == d:
        ret = 1 + endian  # no data blocks in file

    # close file or restore position:
    if fpt is not None:
        if cur_pos is not None:
            fpt.seek(cur_pos, 0)
        else:
            fpt.close()

    return ret


def _read_gcf(file, headonly=False, network='', station='',
              location='', channel_prefix='HH', blockmerge=True,
              cleanoverlap=True, errorret=False, **kwargs):
    """
    Reads a GCF file and returns a :class:`~obspy.core.stream.Stream`

    Only GCF data records are supported. Function supports format as described
    by GCF Reference
    `SWA-RFC-GCFR Issue F, December 2021
    <https://www.guralp.com/apps/ok?doc=GCF_format>`_

    .. warning::
            This function should NOT be called directly, it registers via the
            ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    .. rubric:: Exceptions

        Function will raise ``IOError`` on problems to read file

    .. rubric:: GCF header data

        GCF specific meta data will be set for each
            :class:`~obspy.core.trace.Trace` object in the returned
            :class:`~obspy.core.stream.Stream` object under attribute
            ``stats.gcf``, more precisely the following attributes will be set:

            ``stats.gcf.system_id`` : str
                system_id set in block header
            ``stats.gcf.stream_id`` : str
                6-character stream_id set in block, typically consists of:
                 (4-character unitID) + (1-character orientation code) + (int,
                 tap)
            ``stats.gcf.sys_type`` : int
                system_id type:
                 ``0`` - regular
                 ``1`` - extended
                 ``2`` - double extended
            ``stats.gcf.t_leap`` : bool
                1 if block start at leap second, else 0
            ``stats.gcf.gain`` : int
                variable gain setting (if ``-1`` not used)
            ``stats.gcf.digi`` : int
                digitizer type, combine with ``stats.gcf.sys_type`` to find the
                digitizer according to:
                ==================  ==============  ===========================
                stats.gcf.sys_type  stats.gcf.digi  digitizer
                ==================  ==============  ===========================
                0                   0               unknown (probably DM24 Mk2)
                1                   0               DM24
                1                   1               CD24
                2                   0               Affinity
                2                   1               Minimus
                ==================  ==============  ===========================
            ``stats.gcf.ttl`` : int
                tap-table-lookup to retrieve sequence of decimation filters
                used (see GCF reference linked above for further info)
            ``stats.gcf.blk`` : int
                Lowest block number in gcf file of blocks included in trace
                (count start at 0)
            ``stats.gcf.FIC`` : int
                Forward Integration Constant (i.e. first data value)
            ``stats.gcf.RIC`` : int
                Reverse Integration Constant (ideally last data value)
            ``stats.gcf.stat`` : int
                Status code, available codes are:
                ==== ==========================================================
                code description
                ==== ==========================================================
                -1   not a data block
                0    no detected block issues
                1    stream ID use extended system ID format
                2    stream ID use double-extended system ID format
                3    unknown compression
                4    to few/many data samples indicated in block header
                5    start of first data sample is negative
                9    failure to decode header values
                10   failure to decode data values (last data != RIC, may be
                     due to bad RIC!!!)
                11   first first difference != 0 (may not be critical!!!)
                21   status codes 10 + 11
                ==== ==========================================================

    .. rubric:: Example

        >>> from obspy import read
        >>> st = read("/path/to/20160603_1955n.gcf", format="GCF")

    :type file: str or :class:`~pathlib.Path` or
        file-like object( e.g. `ByesIO`)
    :param file: path to GCF file to read
    :type headonly: bool, optional
    :param headonly: if True only read block headers
    :type network: str, optional
    :param network: network code to use
    :type station: str, optional
    :param station: station code to use, if not specified unitID in gcf header
       will be used (first 4 characters in stream_id)
    :type location: str, optional
    :param location: location code to use
    :type channel_prefix: str, optional
    :param channel_prefix: 2-character channel prefix to use (e.g. ``"HH"`` for
        a high-gain seismometer sampled at 200 Hz)
    :type blockmerge: bool, optional
    :param blockmerge: if True merge blocks if aligned
    :type cleanoverlap: bool, optional
    :param cleanoverlap: ignored if blockmerge is False, if True, remove
        overlapping identical data prior to merging blocks, if False
        overlapping blocks will not be merged even if data in overlap is
        identical
    :type errorret: bool, optional
    :param errorret: if ``True`` block and header issues will be set in
        trace.stats.gcf.stat for each :class:`~obspy.core.trace.Trace`
        object in returned :class:`~obspy.core.stream.Stream` object, else
        function will raise raise an ``IOError`` with an appropriate
        error message
    :rtype: :class:`~obspy.core.stream.Stream`
    :returns: Stream object
    """
    # Load shared library
    gcf_io = _load_cdll("gcf")
    # declare function argument and return types
    gcf_io.free_GcfFile.argtypes = [ctypes.POINTER(_GcfFile)]
    gcf_io.free_GcfFile.restype = None

    # set reader mode
    if headonly:
        if not blockmerge:
            mode = -2
        else:
            mode = -1
    else:
        if not blockmerge:
            mode = 2
        elif not cleanoverlap:
            mode = 1
        else:
            mode = 0

    # read file
    obj = _GcfFile()
    # b_filename = filename.encode('utf-8')
    # ret = gcf_io.read_gcf(fid, obj, mode)
    ret = _fill_gcf_from_file(file, obj, mode)

    if isinstance(file, IOBase):
        if ret:
            raise IOError("failed to read GCF data (error code %d)" % ret)
    else:
        filename = str(file)
        if ret == -1:
            if os.path.isfile(filename):
                raise IOError("cannot open file %s in read mode (missing "
                              "read permissions for user?)" % filename)
            else:
                raise IOError("%s is not a file" % filename)
        elif ret == 1:
            raise IOError("file %s is not a GCF data file" % filename)
        elif ret:
            raise IOError("failed to read file %s (unknown error code %d "
                          "returned)" % (filename, ret))

    err_msg = ""
    # So far so good, set up trace objects
    traces = []
    for i in range(int(obj.n_seg)):
        if obj.seg[i].err and not errorret:
            err_msg = _ERRORMSG_READ.get(
                obj.seg[i].err,
                "unknown error code (%d) set for data block %%d" %
                (obj.seg[i].err)
                ) % (i+1)
            break
        chn = obj.seg[i].streamID.decode('utf-8')[-2]
        channel = channel_prefix + chn.upper()
        if headonly or obj.seg[i].n_data <= 0:
            data = None
        else:
            data = np.array(obj.seg[i].data[:obj.seg[i].n_data],
                            dtype=np.int32)
        sps = obj.seg[i].sps*1./obj.seg[i].sps_denom
        dt = obj.seg[i].t_leap
        start = obj.seg[i].start+obj.seg[i].t_numerator * \
            1./obj.seg[i].t_denominator-dt
        def_stat = obj.seg[i].streamID[:4].decode('utf-8')
        stats = {
            "network": network,
            "station": station if station else def_stat,
            "location": location,
            "channel": channel,
            "sampling_rate": sps,
            "starttime": UTCDateTime(start),
            "npts": obj.seg[i].n_data,
            "gcf": AttribDict({
                "system_id": obj.seg[i].systemID.decode('utf-8'),
                "stream_id": obj.seg[i].streamID.decode('utf-8'),
                "sys_type": obj.seg[i].sysType,
                "t_leap": True if obj.seg[i].t_leap else False,
                "gain": obj.seg[i].gain,
                "digi": obj.seg[i].digi,
                "ttl": obj.seg[i].ttl,
                "blk": obj.seg[i].blk,
                "FIC": obj.seg[i].FIC,
                "RIC": obj.seg[i].RIC,
                "stat": obj.seg[i].err
            })
        }
        if data is not None:
            traces.append(Trace(data=data, header=stats))
        else:
            traces.append(Trace(header=stats))

    # Set up the stream object
    if not err_msg:
        stream = Stream(traces=traces)

    # Free memory allocated by the C-function
    gcf_io.free_GcfFile(obj)

    if err_msg:
        raise IOError(err_msg)
    return stream


def _write_gcf(stream, filename, stream_id=None, system_id=None, is_leap=False,
               gain=None, ttl=None, digi=None, sys_type=None, misalign=0.1,
               **kwargs):
    """
    Writes a :class:`~obspy.core.stream.Stream` or a
    :class:`~obspy.core.trace.Trace` to a GCF file

    Only GCF data records are supported. Function supports format as described
    by `GCF Reference <https://www.guralp.com/apps/ok?doc=GCF_format>`_
    SWA-RFC-GCFR Issue F, December 2021.

    .. warning::

            This function should NOT be called directly, it registers via the
            ObsPy :func:`~obspy.core.stream.write` function, call this
            instead.

    .. note::

        Sampling rate is restricted, most sampling rates between 1-250 samples
        per second (minus a few exceptions)are supported, for greater or lower
        sampling rates format support can be checked with function
        :func:`compatible_sps`.

        First data sample in each trace may only be sampled at non-integer
        second if sampling rate > 250. For sampling rates > 250 first data
        sample in each trace must start at an integer nominator of the
        denominator associated with the ampling rate. e.g. for a sampling rate
        of 1000 samples per second the associated denominator is 4 hence first
        data sample must be sampled at either ss.00 (0/4), ss.25 (1/4), ss.50
        (2/4), or ss.75 (3/4). Use function :func:`get_time_denominator` to get
        the associated denominator.

        The GCF format is only guaranteed to support 32-bit signed integer
        values.  While data with values out of range may be properly stored in
        the GCF format (if first and last data sample can be represented as a
        32-bit signed integer as well as all first difference values of the
        data vector) the current implementaion only permitts input data to be
        representable as a 32-bit signed integer. If input waveforms cannot be
        representable as 32-bit signed integers they will be clipped at
        -2,147,483,648 and 2,147,483,647

    .. rubric:: Exceptions

        ``TypeError`` will be raised upon unsupported input, ``IOError`` will
        be raised upon failure to output file

    .. rubric:: Example

        >>> st.write('GCF-filename.gcf', format='GCF')  # doctest: +SKIP

    :type stream: :class:`~obspy.core.stream.Stream` or
                  :class:`~obspy.core.trace.Trace`
    :param stream: waveform to write to file,
    :type filename: str
    :param filename: path to GCF file to read
    :type stream_id: str
    :param stream_id: 6-character stream ID, first 4 characters (the unit ID)
        should be in the range 0-9, A-Z fifth character should be the
        orientation code (e.g. Z, N, E) and the last character an integer
        yielding the tap the data were output from in the digitizer. If not
        specified stream_id will be taken from ``stats.gcf.stream_id`` for each
        :class:`~obspy.core.trace.Trace` object in the input
        :class:`~obspy.core.stream.Stream` object if present else stream_id
        will be constructed as:
        ``stats.station.ljust(4, x)[:4]+stats.channel[-1]+'0'``
    :type system_id: str
    :param system_id: 4- to 6-character SysID (4 if ``sys_type`` is 2, 5 if
        ``sys_type`` is 1 and 6 if ``sys_type`` is 0). If not specified
       system_id  will be taken from ``stats.gcf.system_id`` for each
        :class:`~obspy.core.trace.Trace` object in the input
        :class:`~obspy.core.stream.Stream` object and if present else system_id
        will be extracted from ``stream_id`` starting from first character.
    :type is_leap: int
    :param is_leap: should be set to ``True`` if first sample is sampled at an
        integer second that is a leap second. If not specified value will be
        taken from ``stats.gcf.t_leap`` for each
        :class:`~obspy.core.trace.Trace` object in the input
        :class:`~obspy.core.stream.Stream` if present else the default value
        is ``0``
    :type gain: int
    :param gain: digitizer gain (permitted values are 2^n; n <= 6, and -1 if
        combined with ``sys_type`` = 0), if not specified gain will be taken
        from ``stats.gcf.gain`` for each :class:`~obspy.core.trace.Trace`
        object in the input  :class:`~obspy.core.stream.Stream` object if
        present else the default value is ``0``.
    :type ttl: int
    :param ttl: tap-table-lookup reference. if not specified ttl will be
        taken from t``stats.gcf.gain`` for each
        :class:`~obspy.core.trace.Trace` object in the input
        :class:`~obspy.core.stream.Stream` object if present else the default
        value is ``0``
    :type digi: int
    :param digi: digitizer type (permitted values are 0 or 1), if not
        specified type will be taken from ``stats.gcf.digi`` for each
        :class:`~obspy.core.trace.Trace` object in the input
        :class:`~obspy.core.stream.Stream` object if present else the default
        value is ``0``
    :type sys_type: int
    :param sys_type: system_id type (permitted values are 0, 1, or 2), if not
        specified sys_type will be taken from ``stats.gcf.sys_type`` for each
        :class:`~obspy.core.trace.Trace` object in the input
        :class:`~obspy.core.stream.Stream` object if present else the default
        value is ``0``
    :type misalign: float
    :param misalign: fraction of a sampling interval (permitted range is
        0-0.5) of tolerated misalignment of starttime. If not specified
        default value is ``0.1``
    """
    # avoid altering input object
    stream = stream.copy()

    # Make sure we have a Stream or Trace object
    if not isinstance(stream, Stream):
        if not isinstance(stream, Trace):
            raise TypeError("write requires either a Stream or a Trace object")
        else:
            stream = Stream(traces=stream)

    # Check output path
    if os.path.exists(filename) and not os.path.isfile(filename):
        IOError("path %s exists but is not a file object" % (filename))
    d, f = os.path.split(filename)
    if d and not os.path.exists(d):
        try:
            os.makedirs(d)
        except Exception:
            raise IOError("failed to create output directorie(s) %s" % (d))

    # check misalign parameter for supported range
    if misalign < 0 or misalign > 0.5:
        raise TypeError("argument misalign out of range (permitted range "
                        "0 - 0.5)")

    ret = 0
    # Load shared library
    gcf_io = _load_cdll("gcf")

    # declare function argument and return types
    gcf_io.write_gcf.argtypes = [ctypes.c_char_p, ctypes.POINTER(_GcfFile)]
    gcf_io.write_gcf.restype = ctypes.c_int

    # prepare the GcfFile object
    obj = _GcfFile()
    obj.n_seg = len(stream)
    obj.seg = (_GcfSeg * obj.n_seg)()
    obj.n_alloc = obj.n_seg
    ctypes.cast(obj.seg, ctypes.POINTER(_GcfSeg))

    for i, trace in enumerate(stream):
        gcf_stats = trace.stats.gcf if hasattr(trace.stats, 'gcf') \
            else None

        # Input checks, all of these are checked in the C-writer function
        #  but still duplicated here for clarity and ease of formatting
        #  proper exception messages

        # Check data type
        if not isinstance(trace.data[0], np.int32):
            if np.any(trace.data > _GCF_MAX) or \
                    np.any(trace.data < _GCF_MIN):
                msg = ('some samples were exceeding the data range that '
                       'GCF format can store (signed 32-bit integer), so '
                       'the data written to file will be clipped')
                warnings.warn(msg)
            trace.data = trace.data.clip(_GCF_MIN, _GCF_MAX)
            data = np.require(trace.data, dtype=np.int32,
                              requirements=['C_CONTIGUOUS'])
            if not np.all(data == trace.data):
                msg = ('data was downcast to int32 and data was altered '
                       'by this operation')
                warnings.warn(msg)
            trace.data = data
        else:
            trace.data = np.ascontiguousarray(trace.data)

        # Check sampling rate
        sps_denom = 1
        if not compatible_sps(trace.stats.sampling_rate):
            sps = ".3f" % (trace.stats.sampling_rate) if \
                           trace.stats.sampling_rate < 1 else \
                           "%d" % (trace.stats.sampling_rate)
            raise ValueError("trace sampling rate, %s, in trace %d not "
                             "supported in GCF format" % (sps, i+1))
        else:
            sps = trace.stats.sampling_rate
            if sps < 1:
                if sps < 0.12:
                    sps_denom = 10
                elif sps < 0.19:
                    sps_denom = 8
                elif sps < 0.24:
                    sps_denom = 5
                elif sps < 0.49:
                    sps_denom = 4
                else:
                    sps_denom = 2
                sps = 1
            sps = int(sps)

        # check start time
        sec = trace.stats.starttime.timestamp
        t_numerator = 0
        if sps <= 250:
            # fractional start time not supported, allow for rounding of
            #  up 10% of the sampling interval
            # hmm is this enough for sampling rates < 1 Hz
            start = int(round(sec))
            if abs(start-sec) > 1./sps*misalign:
                raise ValueError("fractional start time not supported for "
                                 "sampling rates <= 250 Hz (in trace %d)" %
                                 (i+1))
        else:
            # fractional start time supported but restricted, allow for
            # 10% misalignment
            start = int(math.floor(sec))
            dt = sec-start
            t_denom = _SPS_MAP[sps]
            numerator = dt*t_denom
            t_numerator = int(round(numerator))
            if abs(numerator-t_numerator)/t_denom > misalign/sps:
                raise ValueError("start time in trace %d not aligned with "
                                 "supported (fractional) start time" %
                                 (i+1))

        # Check if is_leap is set else set
        use_is_leap = 0
        if is_leap is not None:
            use_is_leap = 1 if is_leap else 0
        elif hasattr(gcf_stats, "is_leap"):
            use_is_leap = 1 if gcf_stats.is_leap else 0

        # Check if gain is set else set
        use_gain = 0
        gain_in_stats = False
        if gain is not None:
            use_gain = gain
        elif hasattr(gcf_stats, "gain"):
            gain_in_stats = True
            use_gain = gcf_stats.gain
        if use_gain not in _VALID_GAIN:
            if gain is None:
                gains = ', '.join(["%d" % g for g in _VALID_GAIN])
                raise ValueError("bad value on stats.gcf.gain, %s, in "
                                 "stats.gcf in trace %d (permitted values:"
                                 " %s)" % (use_gain, i+1, gains))
            else:
                gains = ', '.join(["%d" % g for g in _VALID_GAIN])
                raise ValueError("bad value on argument gain, %s "
                                 "(permitted values: %s)" %
                                 (use_gain, gains))

        # Check if ttl is set else set
        use_ttl = 0
        if ttl is not None:
            use_ttl = ttl
        elif hasattr(gcf_stats, "ttl"):
            use_ttl = gcf_stats.ttl

        # Check if type is set else set
        use_digi = 0
        if digi is not None:
            use_digi = digi
        elif hasattr(gcf_stats, "digi"):
            use_digi = gcf_stats.digi
        if use_digi not in [0, 1]:
            if digi is None:
                raise ValueError("bad value on stats.gcf.digi, %s, in "
                                 "trace %d (permitted values: 0, 1)" %
                                 (use_digi, i+1))
            else:
                raise ValueError("bad value on argument digi, %s "
                                 "(permitted values: 0, 1)" % (use_digi))

        # Check if sys_type is set else set
        use_sys_type = 0
        st_in_stats = False
        if sys_type is not None:
            use_sys_type = sys_type
        elif hasattr(gcf_stats, "sys_type"):
            st_in_stats = True
            use_sys_type = gcf_stats.sys_type
        if use_sys_type not in [0, 1, 2]:
            if digi is None:
                raise ValueError("bad value on stats.gcf.digi, %s, in "
                                 "trace %d (permitted values: 0, 1, 2)"
                                 % (use_sys_type, i+1))
            else:
                raise ValueError("bad value on argument digi, %s, "
                                 "in trace %d (permitted values: 0, 1, 2)"
                                 % (use_sys_type, i+1))
        elif use_sys_type != 0 and use_gain == -1:
            if gain_in_stats and st_in_stats:
                raise ValueError(
                    "value -1 on stats.gcf.gain may only be "
                    "combined with value 0 on "
                    "stats.gcf.sys_type (in trace %s)" % (i+1))
            elif gain_in_stats:
                raise ValueError("value -1 on stats.gcf.gain may only be "
                                 "combined with value 0 on argument "
                                 "sys_type (in trace %s)" % (i+1))
            elif st_in_stats:
                raise ValueError(
                    "value -1 on argument gain may only be "
                    "combined with value 0 on "
                    "stats.gcf.sys_type (in trace %s)" % (i+1))
            else:
                raise ValueError("value -1 on argument gain may only be "
                                 "combined with value 0 on argument "
                                 "sys_type")

        # Check if stream_id is set else build
        if stream_id is not None:
            use_stream_id = stream_id
        elif gcf_stats is not None and hasattr(gcf_stats, "stream_id"):
            use_stream_id = gcf_stats.stream_id.upper()
        else:
            use_stream_id = (trace.stats.station.ljust(4, "X")[:4] +
                             (trace.stats.channel[-1]
                              if trace.stats.channel
                              else 'X')+'0').upper()
        if len(use_stream_id) != 6:
            if stream_id is None:
                raise ValueError("bad value on stats.gcf.stream_id, %s, "
                                 "in trace %d (must be 6-character long)" %
                                 (use_stream_id, i+1))
            else:
                raise ValueError(
                    "bad value on argument stream_id, %s (must"
                    " be 6-character long)" % (use_stream_id))

        # Check if system_id is set else set
        len_system_id = 6 if use_sys_type == 0 \
            else (5 if use_sys_type == 1 else 4)
        use_system_id = use_stream_id[:len_system_id]
        si_in_stats = False
        if system_id is not None:
            use_system_id = system_id.upper()
        elif gcf_stats is not None and hasattr(gcf_stats, "system_id"):
            si_in_stats = True
            use_system_id = gcf_stats.system_id.upper()
        if len(use_system_id) > len_system_id:
            if st_in_stats and si_in_stats:
                raise ValueError("stats.gcf.system_id not compatible with "
                                 "stats.gcf.sys_type in trace %d (max "
                                 "length %d characters )" %
                                 (i+1, len_system_id))
            elif st_in_stats:
                raise ValueError("argument system_id not compatible with "
                                 "stats.gcf.sys_type in trace %d (max "
                                 "length %d characters )" %
                                 (i+1, len_system_id))
            elif si_in_stats:
                raise ValueError("stats.gcf.system_id not compatible with "
                                 "argument sys_type in trace %d (max length"
                                 " %d characters )" % (i+1, len_system_id))
            else:
                raise ValueError("argument system_id not compatible with "
                                 "argument sys_type (max length %d "
                                 "characters )" % (len_system_id))

        # Populate segment
        tr = _GcfSeg()
        tr.streamID = use_stream_id.encode('utf-8')
        tr.systemID = use_system_id.encode('utf-8')
        tr.start = start
        tr.t_numerator = t_numerator
        tr.t_leap = use_is_leap
        tr.gain = use_gain
        tr.sysType = use_sys_type
        tr.digi = use_digi
        tr.ttl = use_ttl
        tr.sps = sps
        tr.sps_denom = sps_denom
        tr.n_data = trace.data.size
        tr.n_alloc = trace.data.size
        tr.data = trace.data.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))

        # Add segment to file object
        obj.seg[i] = tr

    # Write to file
    ret = gcf_io.write_gcf(filename.encode('utf-8'), obj)
    if ret == 0:
        pass
    elif ret in _ERROR_WRITE:
        raise _ERROR_WRITE['ret']
    else:
        raise IOError("unknown error code %d" % (ret))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
A command-line program that runs all ObsPy tests.

For documentation see file obspy/scripts/runtests.py.
"""
from obspy.scripts.runtests import main


if __name__ == "__main__":
    # It is not possible to add the code of main directly to here.
    # This script is automatically installed with name obspy-runtests by
    # setup.py to the Scripts or bin directory of your Python distribution
    # setup.py needs a function to which it's scripts can be linked.
    main()



# -*- coding: utf-8 -*-
"""
obspy.core.event.base - Classes for handling event metadata
===========================================================
This module provides a class hierarchy to consistently handle event metadata.
This class hierarchy is closely modelled after the de-facto standard format
`QuakeML <https://quake.ethz.ch/quakeml/>`_.

.. figure:: /_images/Event.png

.. note::

    For handling additional information not covered by the QuakeML standard and
    how to output it to QuakeML see the :ref:`ObsPy Tutorial <quakeml-extra>`.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import copy
import warnings

import numpy as np

from obspy.core.event.resourceid import ResourceIdentifier
from obspy.core.event.header import DataUsedWaveType, ATTRIBUTE_HAS_ERRORS
from obspy.core.utcdatetime import UTCDateTime
from obspy.core.util import AttribDict


class QuantityError(AttribDict):
    """
    Uncertainty information for a physical quantity.

    :type uncertainty: float
    :param uncertainty: Uncertainty as the absolute value of symmetric
        deviation from the main value.
    :type lower_uncertainty: float
    :param lower_uncertainty: Uncertainty as the absolute value of deviation
        from the main value towards smaller values.
    :type upper_uncertainty: float
    :param upper_uncertainty: Uncertainty as the absolute value of deviation
        from the main value towards larger values.
    :type confidence_level: float
    :param confidence_level: Confidence level of the uncertainty, given in
        percent (0-100).
    """
    defaults = {"uncertainty": None, "lower_uncertainty": None,
                "upper_uncertainty": None, "confidence_level": None}
    warn_on_non_default_key = True

    def __init__(self, uncertainty=None, lower_uncertainty=None,
                 upper_uncertainty=None, confidence_level=None):
        super(QuantityError, self).__init__()
        self.uncertainty = uncertainty
        self.lower_uncertainty = lower_uncertainty
        self.upper_uncertainty = upper_uncertainty
        self.confidence_level = confidence_level

    def __bool__(self):
        """
        Boolean testing for QuantityError.

        QuantityError evaluates ``True`` if any of the default fields is not
        ``None``. Setting non default fields raises also an UserWarning which
        is the reason we have to skip those lines in the doctest below.

        >>> err = QuantityError()
        >>> bool(err)
        False
        >>> err.custom_field = "spam"  # doctest: +SKIP
        >>> bool(err)
        False
        >>> err.uncertainty = 0.05
        >>> bool(err)
        True
        >>> del err.custom_field  # doctest: +SKIP
        >>> bool(err)
        True
        """
        return any([getattr(self, key) is not None for key in self.defaults])

    def __eq__(self, other):
        if other is None and not bool(self):
            return True
        return super(QuantityError, self).__eq__(other)


def _bool(value):
    """
    A custom bool() implementation that returns
    True for any value (including zero) of int and float,
    and for (empty) strings.
    """
    if value == 0 or isinstance(value, str):
        return True
    return bool(value)


def _event_type_class_factory(class_name, class_attributes=[],
                              class_contains=[]):
    """
    Class factory to unify the creation of all the types needed for the event
    handling in ObsPy.

    The types oftentimes share attributes and setting them manually every time
    is cumbersome, error-prone and hard to do consistently. The classes created
    with this method will inherit from
    :class:`~obspy.core.util.attribdict.AttribDict`.

    Usage to create a new class type:

    The created class will assure that any given (key, type) attribute pairs
    will always be of the given type and will attempt to convert any given
    value to the correct type and raise an error otherwise. This happens to
    values given during initialization as well as values set when the object
    has already been created. A useful type is Enum if you want to restrict
    the acceptable values.

        >>> from obspy.core.util import Enum
        >>> MyEnum = Enum(["a", "b", "c"])
        >>> class_attributes = [ \
                ("resource_id", ResourceIdentifier), \
                ("creation_info", CreationInfo), \
                ("some_letters", MyEnum), \
                ("some_error_quantity", float, ATTRIBUTE_HAS_ERRORS), \
                ("description", str)]

    Furthermore the class can contain lists of other objects. There is not much
    to it so far. Giving the name of the created class is mandatory.

        >>> class_contains = ["comments"]
        >>> TestEventClass = _event_type_class_factory("TestEventClass", \
                class_attributes=class_attributes, \
                class_contains=class_contains)
        >>> assert(TestEventClass.__name__ == "TestEventClass")

    Now the new class type can be used.

        >>> test_event = TestEventClass(resource_id="event/123456", \
                creation_info={"author": "obspy.org", "version": "0.1"})

    All given arguments will be converted to the right type upon setting them.

        >>> test_event.resource_id
        ResourceIdentifier(id="event/123456")
        >>> print(test_event.creation_info)
        CreationInfo(author='obspy.org', version='0.1')

    All others will be set to None.

        >>> assert(test_event.description is None)
        >>> assert(test_event.some_letters is None)

    If the resource_id attribute of the created class type is set, the object
    the ResourceIdentifier refers to will be the class instance.

        >>> assert(id(test_event) == \
            id(test_event.resource_id.get_referred_object()))

    They can be set later and will be converted to the appropriate type if
    possible.

        >>> test_event.description = 1
        >>> assert(test_event.description == "1")

    Trying to set with an inappropriate value will raise an error.

        >>> test_event.some_letters = "d" # doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        ValueError: Setting attribute "some_letters" failed. ...

    If you pass ``ATTRIBUTE_HAS_ERRORS`` as the third tuple item for the
    class_attributes, a error (type
    :class:`~obspy.core.event.base.QuantityError`) will be be created that will
    be named like the attribute with "_errors" appended.

        >>> assert(hasattr(test_event, "some_error_quantity_errors"))
        >>> test_event.some_error_quantity_errors  # doctest: +ELLIPSIS
        QuantityError(...)
    """
    class AbstractEventType(AttribDict):
        # Keep the class attributes in a class level list for a manual property
        # implementation that works when inheriting from AttribDict.
        _properties = []
        for item in class_attributes:
            _properties.append((item[0], item[1]))
            if len(item) == 3 and item[2] == ATTRIBUTE_HAS_ERRORS:
                _properties.append((item[0] + "_errors", QuantityError))
        _property_keys = [_i[0] for _i in _properties]
        _property_dict = {}
        for key, value in _properties:
            _property_dict[key] = value
        _containers = class_contains
        warn_on_non_default_key = True
        defaults = dict.fromkeys(class_contains, [])
        defaults.update(dict.fromkeys(_property_keys, None))
        do_not_warn_on = ["extra"]

        def __init__(self, *args, **kwargs):
            # Make sure the args work as expected. Therefore any specified
            # arg will overwrite a potential kwarg, e.g. arg at position 0 will
            # overwrite kwargs class_attributes[0].
            for _i, item in enumerate(args):
                # Use the class_attributes list here because it is not yet
                # polluted be the error quantities.
                kwargs[class_attributes[_i][0]] = item
            # Set all property values to None or the kwarg value.
            for key, _ in self._properties:
                value = kwargs.get(key, None)
                # special handling for resource id
                if key == "resource_id":
                    if kwargs.get("force_resource_id", False):
                        if value is None:
                            value = ResourceIdentifier()
                setattr(self, key, value)
            # Containers currently are simple lists.
            for name in self._containers:
                setattr(self, name, list(kwargs.get(name, [])))
            # All errors are QuantityError. If they are not set yet, set them
            # now.
            for key, _ in self._properties:
                if key.endswith("_errors") and getattr(self, key) is None:
                    setattr(self, key, QuantityError())

        def clear(self):
            """
            Clear the class
            :return:
            """
            super(AbstractEventType, self).clear()
            self.__init__(force_resource_id=False)

        def __str__(self, force_one_line=False):
            """
            Fairly extensive in an attempt to cover several use cases. It is
            always possible to change it in the child class.
            """
            # Get the attribute and containers that are to be printed. Only not
            # None attributes and non-error attributes are printed. The errors
            # will appear behind the actual value.
            # We use custom _bool() for testing getattr() since we want to
            # print int and float values that are equal to zero and empty
            # strings.
            attributes = [_i for _i in self._property_keys if not
                          _i.endswith("_errors") and _bool(getattr(self, _i))]
            containers = [_i for _i in self._containers if
                          _bool(getattr(self, _i))]

            # Get the longest attribute/container name to print all of them
            # nicely aligned.
            max_length = max(max([len(_i) for _i in attributes])
                             if attributes else 0,
                             max([len(_i) for _i in containers])
                             if containers else 0) + 1

            ret_str = self.__class__.__name__

            # Case 1: Empty object.
            if not attributes and not containers:
                return ret_str + "()"

            def get_value_repr(key):
                value = getattr(self, key)
                if isinstance(value, str):
                    value = value
                repr_str = value.__repr__()
                # Print any associated errors.
                error_key = key + "_errors"
                if self.get(error_key, False):
                    err_items = sorted(getattr(self, error_key).items())
                    repr_str += " [%s]" % ', '.join(
                        sorted([str(k) + "=" + str(v) for k, v in err_items
                                if v is not None]))
                return repr_str

            # Case 2: Short representation for small objects. Will just print a
            # single line.
            if len(attributes) <= 3 and not containers or\
               force_one_line:
                att_strs = ["%s=%s" % (_i, get_value_repr(_i))
                            for _i in attributes if _bool(getattr(self, _i))]
                ret_str += "(%s)" % ", ".join(att_strs)
                return ret_str

            # Case 3: Verbose string representation for large object.
            if attributes:
                format_str = "%" + str(max_length) + "s: %s"
                att_strs = [format_str % (_i, get_value_repr(_i))
                            for _i in attributes if _bool(getattr(self, _i))]
                ret_str += "\n\t" + "\n\t".join(att_strs)

            # For the containers just print the number of elements in each.
            if containers:
                # Print delimiter only if there are attributes.
                if attributes:
                    ret_str += '\n\t' + '---------'.rjust(max_length + 5)
                element_str = "%" + str(max_length) + "s: %i Elements"
                ret_str += "\n\t" + \
                    "\n\t".join(
                        [element_str % (_i, len(getattr(self, _i)))
                         for _i in containers])
            return ret_str

        def _repr_pretty_(self, p, cycle):
            p.text(str(self))

        def copy(self):
            return copy.deepcopy(self)

        def __repr__(self):
            return self.__str__(force_one_line=True)

        def __bool__(self):
            # We use custom _bool() for testing getattr() since we want
            # zero valued int and float and empty string attributes to be True.
            if any([_bool(getattr(self, _i))
                    for _i in self._property_keys + self._containers]):
                return True
            return False

        def __eq__(self, other):
            """
            Two instances are considered equal if all attributes and all lists
            are identical.
            """
            # Looping should be quicker on average than a list comprehension
            # because only the first non-equal attribute will already return.
            for attrib in self._property_keys:
                if not hasattr(other, attrib) or \
                   (getattr(self, attrib) != getattr(other, attrib)):
                    return False
            for container in self._containers:
                if not hasattr(other, container) or \
                   (getattr(self, container) != getattr(other, container)):
                    return False
            return True

        def __ne__(self, other):
            return not self.__eq__(other)

        def __setattr__(self, name, value):
            """
            Custom property implementation that works if the class is
            inheriting from AttribDict.
            """
            # avoid type casting of 'extra' attribute, to make it possible to
            # control ordering of extra tags by using an OrderedDict for
            # 'extra'.
            if name == 'extra':
                dict.__setattr__(self, name, value)
                return
            # Pass to the parent method if not a custom property.
            if name not in self._property_dict.keys():
                AttribDict.__setattr__(self, name, value)
                return
            attrib_type = self._property_dict[name]
            # If the value is None or already the correct type just set it.
            if (value is not None) and (type(value) is not attrib_type):
                # If it is a dict, and the attrib_type is no dict, than all
                # values will be assumed to be keyword arguments.
                if isinstance(value, dict):
                    new_value = attrib_type(**value)
                else:
                    new_value = attrib_type(value)
                if new_value is None:
                    msg = 'Setting attribute "%s" failed. ' % (name)
                    msg += 'Value "%s" could not be converted to type "%s"' % \
                        (str(value), str(attrib_type))
                    raise ValueError(msg)
                value = new_value

            # Make sure all floats are finite - otherwise this is most
            # likely a user error.
            if attrib_type is float and value is not None:
                if not np.isfinite(value):
                    msg = "On %s object: Value '%s' for '%s' is " \
                          "not a finite floating point value." % (
                              type(self).__name__, str(value), name)

                    raise ValueError(msg)

            AttribDict.__setattr__(self, name, value)
            # if value is a resource id bind or unbind the resource_id
            if isinstance(value, ResourceIdentifier):
                if name == "resource_id":  # bind the resource_id to self
                    self.resource_id.set_referred_object(self, warn=False)
                else:  # else unbind to allow event scoping later
                    value._parent_key = None

    class AbstractEventTypeWithResourceID(AbstractEventType):
        def __init__(self, force_resource_id=True, *args, **kwargs):
            kwargs["force_resource_id"] = force_resource_id
            super(AbstractEventTypeWithResourceID, self).__init__(*args,
                                                                  **kwargs)

    if "resource_id" in [item[0] for item in class_attributes]:
        base_class = AbstractEventTypeWithResourceID
    else:
        base_class = AbstractEventType

    # Set the class type name.
    setattr(base_class, "__name__", class_name)
    return base_class


__CreationInfo = _event_type_class_factory(
    "__CreationInfo",
    class_attributes=[("agency_id", str),
                      ("agency_uri", ResourceIdentifier),
                      ("author", str),
                      ("author_uri", ResourceIdentifier),
                      ("creation_time", UTCDateTime),
                      ("version", str)])


class CreationInfo(__CreationInfo):
    """
    CreationInfo is used to describe creation metadata (author, version, and
    creation time) of a resource.

    :type agency_id: str, optional
    :param agency_id: Designation of agency that published a resource.
    :type agency_uri: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param agency_uri: Resource Identifier of the agency that published a
        resource.
    :type author: str, optional
    :param author: Name describing the author of a resource.
    :type author_uri: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param author_uri: Resource Identifier of the author of a resource.
    :type creation_time: :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
    :param creation_time: Time of creation of a resource.
    :type version: str, optional
    :param version: Version string of a resource

    >>> info = CreationInfo(author="obspy.org", version="0.0.1")
    >>> print(info)
    CreationInfo(author='obspy.org', version='0.0.1')

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


__TimeWindow = _event_type_class_factory(
    "__TimeWindow",
    class_attributes=[("begin", float),
                      ("end", float),
                      ("reference", UTCDateTime)])


class TimeWindow(__TimeWindow):
    """
    Describes a time window for amplitude measurements, given by a central
    point in time, and points in time before and after this central point. Both
    points before and after may coincide with the central point.

    :type begin: float
    :param begin: Absolute value of duration of time interval before reference
        point in time window. The value may be zero, but not negative. Unit: s
    :type end: float
    :param end: Absolute value of duration of time interval after reference
        point in time window. The value may be zero, but not negative.  Unit: s
    :type reference: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param reference: Reference point in time ("central" point).

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


__CompositeTime = _event_type_class_factory(
    "__CompositeTime",
    class_attributes=[("year", int, ATTRIBUTE_HAS_ERRORS),
                      ("month", int, ATTRIBUTE_HAS_ERRORS),
                      ("day", int, ATTRIBUTE_HAS_ERRORS),
                      ("hour", int, ATTRIBUTE_HAS_ERRORS),
                      ("minute", int, ATTRIBUTE_HAS_ERRORS),
                      ("second", float, ATTRIBUTE_HAS_ERRORS)])


class CompositeTime(__CompositeTime):
    """
    Focal times differ significantly in their precision. While focal times of
    instrumentally located earthquakes are estimated precisely down to seconds,
    historic events have only incomplete time descriptions. Sometimes, even
    contradictory information about the rupture time exist. The CompositeTime
    type allows for such complex descriptions. If the specification is given
    with no greater accuracy than days (i.e., no time components are given),
    the date refers to local time. However, if time components are given, they
    have to refer to UTC.

    :type year: int
    :param year: Year or range of years of the event's focal time.
    :type year_errors: :class:`~obspy.core.event.base.QuantityError`
    :param year_errors: AttribDict containing error quantities.
    :type month: int
    :param month: Month or range of months of the event’s focal time.
    :type month_errors: :class:`~obspy.core.event.base.QuantityError`
    :param month_errors: AttribDict containing error quantities.
    :type day: int
    :param day: Day or range of days of the event’s focal time.
    :type day_errors: :class:`~obspy.core.event.base.QuantityError`
    :param day_errors: AttribDict containing error quantities.
    :type hour: int
    :param hour: Hour or range of hours of the event’s focal time.
    :type hour_errors: :class:`~obspy.core.event.base.QuantityError`
    :param hour_errors: AttribDict containing error quantities.
    :type minute: int
    :param minute: Minute or range of minutes of the event’s focal time.
    :type minute_errors: :class:`~obspy.core.event.base.QuantityError`
    :param minute_errors: AttribDict containing error quantities.
    :type second: float
    :param second: Second and fraction of seconds or range of seconds with
        fraction of the event’s focal time.
    :type second_errors: :class:`~obspy.core.event.base.QuantityError`
    :param second_errors: AttribDict containing error quantities.

    >>> print(CompositeTime(2011, 1, 1))
    CompositeTime(year=2011, month=1, day=1)
    >>> # Can also be instantiated with the uncertainties.
    >>> print(CompositeTime(year=2011, year_errors={"uncertainty":1}))
    CompositeTime(year=2011 [uncertainty=1])

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


__Comment = _event_type_class_factory(
    "__Comment",
    class_attributes=[("text", str),
                      ("resource_id", ResourceIdentifier),
                      ("creation_info", CreationInfo)])


class Comment(__Comment):
    """
    Comment holds information on comments to a resource as well as author and
    creation time information.

    :type text: str
    :param text: Text of comment.
    :type resource_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param resource_id: Resource identifier of comment.
    :type force_resource_id: bool, optional
    :param force_resource_id: If set to False, the automatic initialization of
        `resource_id` attribute in case it is not specified will be skipped.
    :type creation_info: :class:`~obspy.core.event.base.CreationInfo`, optional
    :param creation_info: Creation info for the comment.

    >>> comment = Comment(text="Some comment")
    >>> print(comment)  # doctest:+ELLIPSIS
    Comment(text='Some comment', resource_id=ResourceIdentifier(...))
    >>> comment = Comment(text="Some comment", force_resource_id=False)
    >>> print(comment)
    Comment(text='Some comment')
    >>> comment.resource_id = "comments/obspy-comment-123456"
    >>> print(comment) # doctest:+ELLIPSIS
    Comment(text='Some comment', resource_id=ResourceIdentifier(...))
    >>> comment.creation_info = {"author": "obspy.org"}
    >>> print(comment.creation_info)
    CreationInfo(author='obspy.org')

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


__WaveformStreamID = _event_type_class_factory(
    "__WaveformStreamID",
    class_attributes=[("network_code", str),
                      ("station_code", str),
                      ("channel_code", str),
                      ("location_code", str),
                      ("resource_uri", ResourceIdentifier)])


class WaveformStreamID(__WaveformStreamID):
    """
    Reference to a stream description in an inventory.

    This is mostly equivalent to the combination of networkCode, stationCode,
    locationCode, and channelCode. However, additional information, e. g.,
    sampling rate, can be referenced by the resourceURI. It is recommended to
    use resourceURI as a flexible, abstract, and unique stream ID that allows
    to describe different processing levels, or resampled/filtered products of
    the same initial stream, without violating the intrinsic meaning of the
    legacy identifiers (network, station, channel, and location codes).
    However, for operation in the context of legacy systems, the classical
    identifier components are supported.

    :type network_code: str
    :param network_code: Network code.
    :type station_code: str
    :param station_code: Station code.
    :type location_code: str, optional
    :param location_code: Location code.
    :type channel_code: str, optional
    :param channel_code: Channel code.
    :type resource_uri:
        :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param resource_uri: Resource identifier for the waveform stream.
    :type seed_string: str, optional
    :param seed_string: Provides an alternative initialization way by passing a
        SEED waveform string in the form network.station.location.channel, e.g.
        BW.FUR..EHZ, which will be used to populate the WaveformStreamID's
        attributes.
        It will only be used if the network, station, location and channel
        keyword argument are ALL None.

    .. rubric:: Example

    >>> # Can be initialized with a SEED string or with individual components.
    >>> stream_id = WaveformStreamID(network_code="BW", station_code="FUR",
    ...                              location_code="", channel_code="EHZ")
    >>> print(stream_id) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    WaveformStreamID
          network_code: 'BW'
          station_code: 'FUR'
          channel_code: 'EHZ'
         location_code: ''
    >>> stream_id = WaveformStreamID(seed_string="BW.FUR..EHZ")
    >>> print(stream_id) # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    WaveformStreamID
          network_code: 'BW'
          station_code: 'FUR'
          channel_code: 'EHZ'
         location_code: ''
    >>> # Can also return the SEED string.
    >>> print(stream_id.get_seed_string())
    BW.FUR..EHZ
    """
    def __init__(self, network_code=None, station_code=None,
                 location_code=None, channel_code=None, resource_uri=None,
                 seed_string=None):
        # Use the seed_string if it is given and everything else is not.
        if (seed_string is not None) and (network_code is None) and \
           (station_code is None) and (location_code is None) and \
           (channel_code is None):
            try:
                network_code, station_code, location_code, channel_code = \
                    seed_string.split('.')
            except ValueError:
                warnings.warn("In WaveformStreamID.__init__(): " +
                              "seed_string was given but could not be parsed")
                pass
            if not any([bool(_i) for _i in [network_code, station_code,
                                            location_code, channel_code]]):
                network_code, station_code, location_code, channel_code = \
                    4 * [None]
        super(WaveformStreamID, self).__init__(network_code=network_code,
                                               station_code=station_code,
                                               location_code=location_code,
                                               channel_code=channel_code,
                                               resource_uri=resource_uri)

    def get_seed_string(self):
        """
        Return the seed string representation.

        The seed string is of the form:
            network.station.location.channel
        """
        return "%s.%s.%s.%s" % (
            self.network_code if self.network_code else "",
            self.station_code if self.station_code else "",
            self.location_code if self.location_code else "",
            self.channel_code if self.channel_code else "")

    id = property(get_seed_string)


__ConfidenceEllipsoid = _event_type_class_factory(
    "__ConfidenceEllipsoid",
    class_attributes=[("semi_major_axis_length", float),
                      ("semi_minor_axis_length", float),
                      ("semi_intermediate_axis_length", float),
                      ("major_axis_plunge", float),
                      ("major_axis_azimuth", float),
                      ("major_axis_rotation", float)])


class ConfidenceEllipsoid(__ConfidenceEllipsoid):
    """
    This class represents a description of the location uncertainty as a
    confidence ellipsoid with arbitrary orientation in space. See the QuakeML
    documentation for the full details

    :param semi_major_axis_length: Largest uncertainty, corresponding to the
        semi-major axis of the confidence ellipsoid. Unit: m
    :param semi_minor_axis_length: Smallest uncertainty, corresponding to the
        semi-minor axis of the confidence ellipsoid. Unit: m
    :param semi_intermediate_axis_length: Uncertainty in direction orthogonal
        to major and minor axes of the confidence ellipsoid. Unit: m
    :param major_axis_plunge: Plunge angle of major axis of confidence
        ellipsoid. Corresponds to Tait-Bryan angle φ. Unit: deg
    :param major_axis_azimuth: Azimuth angle of major axis of confidence
        ellipsoid. Corresponds to Tait-Bryan angle ψ. Unit: deg
    :param major_axis_rotation: This angle describes a rotation about the
        confidence ellipsoid’s major axis which is required to define the
        direction of the ellipsoid’s minor axis. Corresponds to Tait-Bryan
        angle θ.
        Unit: deg

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


__DataUsed = _event_type_class_factory(
    "__DataUsed",
    class_attributes=[("wave_type", DataUsedWaveType),
                      ("station_count", int),
                      ("component_count", int),
                      ("shortest_period", float),
                      ("longest_period", float)])


class DataUsed(__DataUsed):
    """
    The DataUsed class describes the type of data that has been used for a
    moment-tensor inversion.

    :type wave_type: str
    :param wave_type: Type of waveform data.
        See :class:`~obspy.core.event.header.DataUsedWaveType` for allowed
        values.
    :type station_count: int, optional
    :param station_count: Number of stations that have contributed data of the
        type given in wave_type.
    :type component_count: int, optional
    :param component_count: Number of data components of the type given in
        wave_type.
    :type shortest_period: float, optional
    :param shortest_period: Shortest period present in data. Unit: s
    :type longest_period: float, optional
    :param longest_period: Longest period present in data. Unit: s

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Quick and dirty conversion routine from CSS 2.8 to Seismic Handler ASCII format

- expects wfdisc index file as parameter
- processes only first line of wfdisc (no support for multiple streams)
- output written to stdout
- shows plot for inspection (if matplotlib installed)
"""
import struct
import sys
from datetime import datetime


if __name__ == '__main__':
    # read header (assume CSS2.8)
    try:
        # only process first line
        head = list(map(str.strip, open(sys.argv[1]).readlines()[0].split()))
    except Exception:
        sys.exit("cannot read wfdisc file (arg)")

    input = head[15]  # input file name
    # 0 -> timestamp, 1-> milliseconds
    timedata = list(map(int, head[1].split(".")))

    # headers for SH ASCII file
    SH = {
        "STATION": head[2],
        "COMP": head[3][-1].upper(),  # guess from channel naming
        "START": ".".join((
            datetime.fromtimestamp(timedata[0]).strftime("%d-%b-%Y_%H:%M:%S"),
            "%03d" % timedata[1])),
        "DELTA": 1.0 / float(head[5]),
        "LENGTH": int(head[4]),
        "CALIB": float(head[6])
    }

    # binary format (big endian integers)
    fmt = ">" + "i" * SH["LENGTH"]

    # convert binary data
    data = struct.unpack(fmt, open(input, "rb").read(struct.calcsize(fmt)))

    # echo headers
    for header in SH:
        print(": ".join((header, str(SH[header]))))

    # echo data
    for dat in data:
        # CALIB factor
        print("%e" % (dat * SH["CALIB"],))

    # inspection plot
    try:
        import pylab
        pylab.plot([x * SH["DELTA"] for x in range(SH["LENGTH"])],
                   [d * SH["CALIB"] for d in data])
        pylab.show()
    except Exception:
        sys.exit("cannot show plot!")



import obspy


singlechannel = obspy.read('https://examples.obspy.org/COP.BHZ.DK.2009.050')
dt = singlechannel[0].stats.starttime
singlechannel.plot(color='red', tick_rotation=5, tick_format='%I:%M %p',
                   starttime=dt + 60 * 60, endtime=dt + 60 * 60 + 120)



# -*- coding: utf-8 -*-
from obspy.core.util import get_example_file
from obspy.core.util.decorator import map_example_filename


class TestUtilDecorator:
    def test_map_example_filename(self):
        """
        Tests the @map_example_filename decorator
        """
        dummy = "abc"
        example_file = "example.npz"
        path = "/path/to/" + example_file
        path_mapped = get_example_file(example_file)

        def unchanged(a, b="", **kwargs):
            return list(map(str, (a, b, kwargs)))

        @map_example_filename("a")
        def changed1(a, b="", **kwargs):
            return list(map(str, (a, b, kwargs)))
        assert changed1(dummy, dummy) == unchanged(dummy, dummy)
        assert changed1(path, dummy) == unchanged(path_mapped, dummy)
        assert changed1(dummy, path) == unchanged(dummy, path)
        assert changed1(a=path, b=dummy) == unchanged(path_mapped, dummy)
        assert changed1(path, b=dummy) == unchanged(path_mapped, dummy)
        assert changed1(path, b=path, x=path) == \
            unchanged(path_mapped, path, x=path)

        @map_example_filename("b")
        def changed2(a, b="", **kwargs):
            return list(map(str, (a, b, kwargs)))
        assert changed2(dummy, dummy) == unchanged(dummy, dummy)
        assert changed2(path, dummy) == unchanged(path, dummy)
        assert changed2(dummy, path) == unchanged(dummy, path_mapped)
        assert changed2(a=path, b=dummy) == unchanged(path, dummy)
        assert changed2(path, b=path) == unchanged(path, path_mapped)
        assert changed2(path, b=path, x=path) == \
            unchanged(path, path_mapped, x=path)

        @map_example_filename("x")
        def changed3(a, b="", **kwargs):
            return list(map(str, (a, b, kwargs)))
        assert changed3(dummy, dummy) == unchanged(dummy, dummy)
        assert changed3(path, dummy) == unchanged(path, dummy)
        assert changed3(dummy, path) == unchanged(dummy, path)
        assert changed3(a=path, b=dummy) == unchanged(path, dummy)
        assert changed3(path, b=dummy) == unchanged(path, dummy)
        assert changed3(path, b=path, x=path) == \
            unchanged(path, path, x=path_mapped)



# -*- coding: utf-8 -*-
# -----------------------------------------------
# Filename: mopad_wrapper.py
#  Purpose: Wrapper for mopad
#   Author: Tobias Megies, Moritz Beyreuther
#    Email: megies@geophysik.uni-muenchen.de
#
# Copyright (C) 2008-2012 ObsPy Development Team
# -----------------------------------------------
"""
ObsPy wrapper to the *Moment tensor Plotting and Decomposition tool* (MoPaD)
written by Lars Krieger and Sebastian Heimann.

.. seealso:: [Krieger2012]_

.. warning:: The MoPaD wrapper does not yet provide the full functionality of
    MoPaD. Please consider using the command line script ``obspy-mopad`` for
    now if you need the full power of MoPaD.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import numpy as np

from obspy.imaging.beachball import xy2patch
from obspy.imaging.scripts.mopad import BeachBall as mopad_BeachBall
from obspy.imaging.scripts.mopad import MomentTensor as mopad_MomentTensor
from obspy.imaging.scripts.mopad import epsilon


# seems the base system we (gmt) are using is called "USE" in mopad
KWARG_MAP = {
    'size': ['plot_size', 'plot_aux_plot_size'],
    'linewidth': ['plot_nodalline_width', 'plot_outerline_width'],
    'facecolor': ['plot_tension_colour'],
    'edgecolor': ['plot_outerline_colour'],
    'bgcolor': [],
    'alpha': ['plot_total_alpha'],
    'width': [],
    'outfile': ['plot_outfile'],
    'format': ['plot_outfile_format'],
    'nofill': ['plot_only_lines']
}


def _normalize_focmec(fm):
    """
    Improve stability of plots by normalizing the moment tensors. The scale
    does not matter for the beachballs.
    """
    # Only normalize 6 component tensors.
    if len(fm) != 6:
        return fm
    fm = np.array(fm, dtype=np.float64)
    fm /= np.linalg.norm(fm)
    return fm


def beach(fm, linewidth=2, facecolor='b', bgcolor='w', edgecolor='k',
          alpha=1.0, xy=(0, 0), width=200, size=100, nofill=False,
          zorder=100, mopad_basis='USE', axes=None):
    """
    Return a beach ball as a collection which can be connected to an
    current matplotlib axes instance (ax.add_collection). Based on MoPaD.

    S1, D1, and R1, the strike, dip and rake of one of the focal planes, can
    be vectors of multiple focal mechanisms.

    :param fm: Focal mechanism that is either number of mechanisms (NM) by 3
        (strike, dip, and rake) or NM x 6 (M11, M22, M33, M12, M13, M23 - the
        six independent components of the moment tensor, where the coordinate
        system is 1,2,3 = Up,South,East which equals r,theta,phi -
        Harvard/Global CMT convention). The relation to Aki and Richards
        x,y,z equals North,East,Down convention is as follows: Mrr=Mzz,
        Mtt=Mxx, Mpp=Myy, Mrt=Mxz, Mrp=-Myz, Mtp=-Mxy.
        The strike is of the first plane, clockwise relative to north.
        The dip is of the first plane, defined clockwise and perpendicular to
        strike, relative to horizontal such that 0 is horizontal and 90 is
        vertical. The rake is of the first focal plane solution. 90 moves the
        hanging wall up-dip (thrust), 0 moves it in the strike direction
        (left-lateral), -90 moves it down-dip (normal), and 180 moves it
        opposite to strike (right-lateral).
    :param facecolor: Color to use for quadrants of tension; can be a string,
        e.g. ``'r'``, ``'b'`` or three component color vector, [R G B].
        Defaults to ``'b'`` (blue).
    :param bgcolor: The background color. Defaults to ``'w'`` (white).
    :param edgecolor: Color of the edges. Defaults to ``'k'`` (black).
    :param alpha: The alpha level of the beach ball. Defaults to ``1.0``
        (opaque).
    :param xy: Origin position of the beach ball as tuple. Defaults to
        ``(0, 0)``.
    :type width: int
    :param width: Symbol size of beach ball. Defaults to ``200``.
    :param size: Controls the number of interpolation points for the
        curves. Minimum is automatically set to ``100``.
    :param nofill: Do not fill the beach ball, but only plot the planes.
    :param zorder: Set zorder. Artists with lower zorder values are drawn
        first.
    :param mopad_basis: The basis system. Defaults to ``'USE'``. See the
        `Supported Basis Systems`_ section below for a full list of supported
        systems.
    :type axes: :class:`matplotlib.axes.Axes`
    :param axes: Used to make beach balls circular on non-scaled axes. Also
        maintains the aspect ratio when resizing the figure. Will not add
        the returned collection to the axes instance.

    .. rubric:: _`Supported Basis Systems`

    ========= =================== =============================================
    Short     Basis vectors       Usage
    ========= =================== =============================================
    ``'NED'`` North, East, Down   Jost and Herrmann 1989
    ``'USE'`` Up, South, East     Global CMT Catalog, Larson et al. 2010
    ``'XYZ'`` East, North, Up     General formulation, Jost and Herrmann 1989
    ``'RT'``  Radial, Transverse, psmeca (GMT), Wessel and Smith 1999
              Tangential
    ``'NWU'`` North, West, Up     Stein and Wysession 2003
    ========= =================== =============================================
    """
    import matplotlib.collections as mpl_collections
    from matplotlib import patches, transforms
    fm = _normalize_focmec(fm)
    # initialize beachball
    mt = mopad_MomentTensor(fm, system=mopad_basis)
    bb = mopad_BeachBall(mt, npoints=size)
    bb._setup_BB(unit_circle=False)

    # extract the coordinates and colors of the lines
    radius = width / 2.0
    neg_nodalline = bb._nodalline_negative_final_US
    pos_nodalline = bb._nodalline_positive_final_US
    tension_colour = facecolor
    pressure_colour = bgcolor

    if nofill:
        tension_colour = 'none'
        pressure_colour = 'none'

    # based on mopads _setup_plot_US() function
    # collect patches for the selection
    coll = [None, None, None]
    coll[0] = patches.Circle(xy, radius=radius)
    coll[1] = xy2patch(neg_nodalline[0, :], neg_nodalline[1, :], radius, xy)
    coll[2] = xy2patch(pos_nodalline[0, :], pos_nodalline[1, :], radius, xy)

    # set the color of the three parts
    fc = [None, None, None]
    if bb._plot_clr_order > 0:
        fc[0] = pressure_colour
        fc[1] = tension_colour
        fc[2] = tension_colour
        if bb._plot_curve_in_curve != 0:
            fc[0] = tension_colour
            if bb._plot_curve_in_curve < 1:
                fc[1] = pressure_colour
                fc[2] = tension_colour
            else:
                coll = [coll[i] for i in (0, 2, 1)]
                fc[1] = pressure_colour
                fc[2] = tension_colour
    else:
        fc[0] = tension_colour
        fc[1] = pressure_colour
        fc[2] = pressure_colour
        if bb._plot_curve_in_curve != 0:
            fc[0] = pressure_colour
            if bb._plot_curve_in_curve < 1:
                fc[1] = tension_colour
                fc[2] = pressure_colour
            else:
                coll = [coll[i] for i in (0, 2, 1)]
                fc[1] = tension_colour
                fc[2] = pressure_colour

    if bb._pure_isotropic:
        if abs(np.trace(bb._M)) > epsilon:
            # use the circle as the most upper layer
            coll = [coll[0]]
            if bb._plot_clr_order < 0:
                fc = [tension_colour]
            else:
                fc = [pressure_colour]

    # transform the patches to a path collection and set
    # the appropriate attributes
    collection = mpl_collections.PatchCollection(coll, match_original=False)
    collection.set_facecolors(fc)
    # Use the given axes to maintain the aspect ratio of beachballs on figure
    # resize.
    if axes is not None:
        # This is what holds the aspect ratio (but breaks the positioning)
        collection.set_transform(transforms.Affine2D(np.identity(3)))
        # Next is a dirty hack to fix the positioning:
        # 1. Need to bring the all patches to the origin (0, 0).
        for p in collection._paths:
            p.vertices -= xy
        # 2. Then use the offset property of the collection to position the
        # patches
        collection.set_offsets(xy)
        try:
            collection.set_offset_transform(axes.transData)
        except AttributeError:
            # compatibility for matplotlib 3.3 (and maybe 3.4 too?)
            collection._transOffset = axes.transData
    collection.set_edgecolors(edgecolor)
    collection.set_alpha(alpha)
    collection.set_linewidth(linewidth)
    collection.set_zorder(zorder)
    return collection


def beachball(fm, linewidth=2, facecolor='b', bgcolor='w', edgecolor='k',
              alpha=1.0, xy=(0, 0), width=200, size=100, nofill=False,
              zorder=100, mopad_basis='USE', outfile=None, format=None,
              fig=None):
    """
    Draws a beach ball diagram of an earthquake focal mechanism. Based on
    MoPaD.

    S1, D1, and R1, the strike, dip and rake of one of the focal planes, can
    be vectors of multiple focal mechanisms.

    :param fm: Focal mechanism that is either number of mechanisms (NM) by 3
        (strike, dip, and rake) or NM x 6 (M11, M22, M33, M12, M13, M23 - the
        six independent components of the moment tensor, where the coordinate
        system is 1,2,3 = Up,South,East which equals r,theta,phi). The strike
        is of the first plane, clockwise relative to north.
        The dip is of the first plane, defined clockwise and perpendicular to
        strike, relative to horizontal such that 0 is horizontal and 90 is
        vertical. The rake is of the first focal plane solution. 90 moves the
        hanging wall up-dip (thrust), 0 moves it in the strike direction
        (left-lateral), -90 moves it down-dip (normal), and 180 moves it
        opposite to strike (right-lateral).
    :param facecolor: Color to use for quadrants of tension; can be a string,
        e.g. ``'r'``, ``'b'`` or three component color vector, [R G B].
        Defaults to ``'b'`` (blue).
    :param bgcolor: The background color. Defaults to ``'w'`` (white).
    :param edgecolor: Color of the edges. Defaults to ``'k'`` (black).
    :param alpha: The alpha level of the beach ball. Defaults to ``1.0``
        (opaque).
    :param xy: Origin position of the beach ball as tuple. Defaults to
        ``(0, 0)``.
    :type width: int
    :param width: Symbol size of beach ball. Defaults to ``200``.
    :param size: Controls the number of interpolation points for the
        curves. Minimum is automatically set to ``100``.
    :param nofill: Do not fill the beach ball, but only plot the planes.
    :param zorder: Set zorder. Artists with lower zorder values are drawn
        first.
    :param mopad_basis: The basis system. Defaults to ``'USE'``. See the
        `Supported Basis Systems`_ section below for a full list of supported
        systems.
    :param outfile: Output file string. Also used to automatically
        determine the output format. Supported file formats depend on your
        matplotlib backend. Most backends support png, pdf, ps, eps and
        svg. Defaults to ``None``.
    :param format: Format of the graph picture. If no format is given the
        outfile parameter will be used to try to automatically determine
        the output format. If no format is found it defaults to png output.
        If no outfile is specified but a format is, than a binary
        imagestring will be returned.
        Defaults to ``None``.
    :param fig: Give an existing figure instance to plot into. New Figure if
        set to ``None``.

    .. rubric:: _`Supported Basis Systems`

    ========= =================== =============================================
    Short     Basis vectors       Usage
    ========= =================== =============================================
    ``'NED'`` North, East, Down   Jost and Herrmann 1989
    ``'USE'`` Up, South, East     Global CMT Catalog, Larson et al. 2010
    ``'XYZ'`` East, North, Up     General formulation, Jost and Herrmann 1989
    ``'RT'``  Radial, Transverse, psmeca (GMT), Wessel and Smith 1999
              Tangential
    ``'NWU'`` North, West, Up     Stein and Wysession 2003
    ========= =================== =============================================

    .. rubric:: Examples

    (1) Using basis system ``'NED'``.

        >>> from obspy.imaging.mopad_wrapper import beachball
        >>> mt = [1, 2, 3, -4, -5, -10]
        >>> beachball(mt, mopad_basis='NED') #doctest: +SKIP

        .. plot::

            from obspy.imaging.mopad_wrapper import beachball
            mt = [1, 2, 3, -4, -5, -10]
            beachball(mt, mopad_basis='NED')
    """
    fm = _normalize_focmec(fm)
    mopad_kwargs = {}
    loc = locals()
    # map to kwargs used in mopad
    for key in KWARG_MAP:
        value = loc[key]
        for mopad_key in KWARG_MAP[key]:
            mopad_kwargs[mopad_key] = value
    # convert from points to size in cm
    for key in ['plot_aux_plot_size', 'plot_size']:
        # 100.0 is matplotlib's default DPI for savefig
        mopad_kwargs[key] = mopad_kwargs[key] / 100.0 * 2.54
    # use nofill kwarg

    mt = mopad_MomentTensor(fm, system=mopad_basis)
    bb = mopad_BeachBall(mt, npoints=size)

    # show plot in a window
    if outfile is None:
        bb.ploBB(mopad_kwargs)
    # save plot to file
    else:
        # no format specified, parse it from outfile name
        if mopad_kwargs['plot_outfile_format'] is None:
            mopad_kwargs['plot_outfile_format'] = \
                mopad_kwargs['plot_outfile'].split(".")[-1]
        else:
            # append file format if not already at end of outfile
            if not mopad_kwargs['plot_outfile'].endswith(
               mopad_kwargs['plot_outfile_format']):
                mopad_kwargs['plot_outfile'] += "." + \
                    mopad_kwargs['plot_outfile_format']
        bb.save_BB(mopad_kwargs)
